"""The laue module provide helpers functions to work with polychromatic X-ray diffraction.
"""
import numpy as np
from math import cos, sin, tan, atan2, pi
from pymicro.crystal.lattice import HklPlane, Symmetry, HklDirection, HklObject
from pymicro.crystal.microstructure import Orientation
from pymicro.xray.xray_utils import lambda_nm_to_keV, lambda_keV_to_nm
from pymicro.xray.dct import add_to_image
import os


def select_lambda(hkl, orientation, Xu=(1., 0., 0.), verbose=False):
    """
    Compute the wavelength corresponding to the first order reflection
    of a given lattice plane in the specified orientation.

    :param hkl: The given lattice plane.
    :param orientation: The orientation of the crystal lattice.
    :param Xu: The unit vector of the incident X-ray beam (default along the X-axis).
    :param bool verbose: activate verbose mode (default False).
    :returns tuple: A tuple of the wavelength value (keV) and the corresponding Bragg angle (radian).
    """
    (h, k, l) = hkl.miller_indices()
    d_hkl = hkl.interplanar_spacing()
    Gc = hkl.scattering_vector()
    gt = orientation.orientation_matrix().transpose()
    Gs = gt.dot(Gc)
    # compute the theta angle between X and the scattering vector which is theta + pi/2
    theta = np.arccos(np.dot(Xu, Gs / np.linalg.norm(Gs))) - pi / 2
    # find the lambda selected by this hkl plane
    the_energy = lambda_nm_to_keV(2 * d_hkl * np.sin(theta))
    if verbose:
        print('\n** Looking at plane %d%d%d **' % (h, k, l))
        print('scattering vector in laboratory frame = %s' % Gs)
        print('glancing angle between incident beam and the hkl plane is %.1f deg' % (theta * 180 / pi))
        print('corresponding energy is %.1f keV' % the_energy)
    return the_energy, theta


def build_list(lattice=None, max_miller=3, extinction=None, Laue_extinction=False, max_keV=120.):
    hklplanes = []
    indices = range(-max_miller, max_miller + 1)
    for h in indices:
        for k in indices:
            for l in indices:
                if h == k == l == 0:  # skip (0, 0, 0)
                    continue
                if not extinction :
                    hklplanes.append(HklPlane(h, k, l, lattice))
                if extinction == 'FCC':
                    # take plane if only all odd or all even hkl indices
                    if (h % 2 == 0 and k % 2 == 0 and l % 2 == 0) or (h % 2 == 1 and k % 2 == 1 and l % 2 == 1):
                        hklplanes.append(HklPlane(h, k, l, lattice))
                if extinction == 'BCC':
                    # take plane only if the sum indices is even
                    if ((h**2 + k**2 + l**2) % 2 == 0):
                        hklplanes.append(HklPlane(h, k, l, lattice))
    if Laue_extinction is True:
        lam_min = lambda_keV_to_nm(max_keV)
        val = 2. * lattice._lengths[0] / lam_min # lattice have to be cubic !
        print('Limit value is %d' % val)
        for hkl in hklplanes:
            (h, k, l) = hkl.miller_indices()
            test = h ** 2 + k ** 2 + l ** 2
            if val < test: # TODO check the test
                hklplanes.remove(HklPlane(h, k, l, lattice))

    return hklplanes


def compute_ellipsis(orientation, detector, uvw, Xu=(1., 0., 0.), n=101, verbose=False):
    """
    Compute the ellipsis associated with the given zone axis. 
    
    The detector is supposed to be normal to the X-axis, but the incident wave vector can be specified.
    All lattice planes sharing this zone axis will diffract along that ellipse.

    :param orientation: The crystal orientation.
    :param detector: An instance of the Detector2D class.
    :param uvw: An instance of the HklDirection representing the zone axis.
    :param Xu: The unit vector of the incident X-ray beam (default along the X-axis).
    :param int n: number of poits used to define the ellipse.
    :param bool verbose: activate verbose mode (default False).
    :returns data: the (Y, Z) data (in mm unit) representing the ellipsis.
    """
    gt = orientation.orientation_matrix().transpose()
    za = gt.dot(uvw.direction())  # zone axis unit vector in lab frame
    OA = detector.project_along_direction(za)  # vector from the origin to projection of the zone axis onto the detector
    OC = detector.project_along_direction(Xu)  # C is the intersection of the direct beam with the detector
    ON = detector.project_along_direction(detector.w_dir)  # N is the intersection of the X-axis with the detector plane
    CA = OA - OC
    NA = OA - ON
    # psi = np.arccos(np.dot(ZAD/np.linalg.norm(ZAD), Xu))
    psi = atan2(np.linalg.norm(np.cross(za, Xu)), np.dot(za, Xu))  # use cross product instead of arccos
    nu = atan2(np.linalg.norm(np.cross(za, detector.w_dir)), np.dot(za, detector.w_dir))
    e = np.sin(nu) / np.cos(psi)  # ellipis excentricity (general case), e reduces to tan(psi) when the incident beam is normal to the detector (nu = psi)
    '''
    The major axis direction is given by the intersection of the plane defined by ON and OA and by the detector 
    plane. It does not depends on Xu, like the cone angle. If the detector is perpendicular to the X-axis, 
    this direction is given by NA. both N and A are located on the major axis.
    '''
    eta = pi / 2 - atan2(NA[1],
                         NA[2])  # atan2(y, x) compute the arc tangent of y/x considering the sign of both y and x
    (uc, vc) = (detector.ucen - OA[1] / detector.pixel_size, detector.vcen - OA[2] / detector.pixel_size)
    '''
    # wrong formulae from Amoros (The Laue Method)...
    a = ON * np.tan(psi)
    b = ON * np.sin(psi)
    # the proper equation for normal incidence are (Cruickshank 1991):
    a = abs(0.5 * ON * np.tan(2 * psi))
    b = abs(0.5 * ON * np.tan(2 * psi) * np.sqrt(1 - np.tan(psi) ** 2))
    this equation does not hold if the incident beam is not normal with the detector !
    the general equation is:
    2.a / ON = tan(nu) + tan(psi - nu) + sin(psi) / (cos(nu) * cos(psi + nu))
    this reduces when nu = psi to 2.a / ON = np.tan(2.psi)
    '''
    #factor = tan(nu) + tan(psi - nu) + sin(psi) / (cos(nu) * cos(psi + nu))
    #a = 0.5 * np.linalg.norm(ON) * factor
    # a = abs(0.5 * np.linalg.norm(ON) * np.tan(2 * psi))
    # use the factorized version of the factor calculated in the article
    factor = cos(psi) * sin(psi) / (cos(psi + nu) * cos(psi - nu))
    a = np.linalg.norm(ON) * factor
    b = a * np.sqrt(1 - e ** 2)
    if verbose:
        print('angle nu (deg) is %.3f' % (nu * 180 / pi))
        print('angle psi (deg) is %.3f' % (psi * 180 / pi))
        print('angle eta (deg) is %.3f' % (eta * 180 / pi))
        print('direct beam crosses the det plane at %s' % OC)
        print('zone axis crosses the det plane at %s' % OA)
        print('zone axis crosses the detector at (%.3f,%.3f) mm or (%d,%d) pixels' % (OA[1], OA[2], uc, vc))
        print('ellipse eccentricity is %f' % e)
        print('ellipsis major and minor half axes are a=%.3f and b=%.3f' % (a, b))
    # use a parametric curve to plot the ellipsis
    t = np.linspace(0., 2 * pi, n)
    x = a * np.cos(t)
    y = b * np.sin(t)
    data = np.array([x, y])
    # rotate the ellipse
    R = np.array([[np.cos(eta), -np.sin(eta)], [np.sin(eta), np.cos(eta)]])
    data = np.dot(R, data)  # rotate our ellipse
    # move one end of the great axis to the direct beam position
    '''
    NB: in the general case, point C (intersection of the direct beam and the detector) is on the ellipse 
    but not lying on the great axis. Point A and N still lie on the major axis, so the translation can be determined. 
    '''
    NX = np.linalg.norm(ON) * tan(psi - nu)
    if verbose:
        print('****************** NX = %.1f' % NX)
    data[0] += (a - NX) * np.cos(eta)
    data[1] += (a - NX) * np.sin(eta)
    #data[0] += ON[1] + a * np.cos(eta)
    #data[1] += ON[2] + a * np.sin(eta)
    yz_data = np.empty((n, 2), dtype=float)
    for i in range(n):
        yz_data[i, 0] = data[0, i]
        yz_data[i, 1] = data[1, i]
    return yz_data


def diffracted_vector(hkl, orientation, Xu=(1., 0., 0.), min_theta=0.1, use_friedel_pair=True, verbose=False):
    """Compute the diffraction vector for a given reflection of a crystal.

    This method compute the diffraction vector.
    The incident wave vector is along the X axis by default but can be changed by specifying any unit vector.

    :param hkl: an instance of the `HklPlane` class.
    :param orientation: an instance of the `Orientation` class.
    :param Xu: The unit vector of the incident X-ray beam (default along the X-axis).
    :param float min_theta: the minimum considered Bragg angle (in radian).
    :param bool use_friedel_pair: also consider the Friedel pairs of the lattice plane. 
    :param bool verbose: a flag to activate verbose mode.
    :return: the diffraction vector as a numpy array.
    """
    gt = orientation.orientation_matrix().transpose()
    (h, k, l) = hkl.miller_indices()
    # this hkl plane will select a particular lambda
    (the_energy, theta) = select_lambda(hkl, orientation, Xu=Xu, verbose=verbose)
    if the_energy < 0:
        if use_friedel_pair:
            # use the Friedel pair of this lattice plane
            hkl = hkl.friedel_pair()
            (h, k, l) = hkl.miller_indices()
            if verbose:
                print('switched to friedel pair (%d,%d,%d)' % (h, k, l))
            (the_energy, theta) = select_lambda(hkl, orientation, Xu=Xu, verbose=verbose)
        else:
            return None
    assert(theta >= 0)
    if theta < min_theta * pi / 180:  # skip angles < min_theta deg
        return None
    lambda_nm = 1.2398 / the_energy
    X = np.array(Xu) / lambda_nm
    Gs = gt.dot(hkl.scattering_vector())
    if verbose:
        print('bragg angle for %d%d%d reflection is %.1f' % (h, k, l, hkl.bragg_angle(the_energy) * 180 / pi))
    assert (abs(theta - hkl.bragg_angle(the_energy)) < 1e-6)  # verify than theta_bragg is equal to the glancing angle
    # compute diffracted direction
    K = X + Gs
    return K


def diffracted_intensity(hkl, I0=1.0, symbol='Ni', verbose=False):
    '''Compute the diffracted intensity.
    
    This compute a number representing the diffracted intensity within the kinematical approximation. 
    This number includes many correction factors:
    
     * atomic factor
     * structural factor
     * polarisation factor
     * Debye-Waller factor
     * absorption factor
     * ...
    
    :param HklPlane hkl: the hkl Bragg reflection.
    :param float I0: intensity of the incident beam.
    :param str symbol: string representing the considered lattice.
    :param bool verbose: flag to activate verbose mode.
    :return: the diffracted intensity.
    :rtype: float.
    '''
    # TODO find a way to load the atom scattering data from the lattice... and not load it for every reflection!
    path = os.path.dirname(__file__)
    scat_data = np.genfromtxt(os.path.join(path, 'data', '%s_atom_scattering.txt' % symbol))
    # scale the intensity with the atomic scattering factor
    q = 1 / (2 * hkl.interplanar_spacing())
    try:
        index = (scat_data[:, 0] < q).tolist().index(False)
    except ValueError:
        index = -1
    fa = scat_data[index, 1] / scat_data[0, 1]  # should we normalize with respect to Z?

    I = I0 * fa

    if verbose:
        print('q=%f in nm^-1 -> fa=%.3f' % (q, fa))
        print('intensity:', I)
    return I


def compute_Laue_pattern(orientation, detector, hkl_planes=None, Xu=np.array([1., 0., 0.]), use_friedel_pair=False,
                         spectrum=None, spectrum_thr=0., r_spot=5, color_field='constant', inverted=False,
                         show_direct_beam=False, verbose=False):
    """
    Compute a transmission Laue pattern. The data array of the given
    `Detector2d` instance is initialized with the result.
    
    The incident beam is assumed to be along the X axis: (1, 0, 0) but can be changed to any direction. 
    The crystal can have any orientation using an instance of the `Orientation` class. 
    The `Detector2d` instance holds all the geometry (detector size and position).

    A parameter controls the meaning of the values in the diffraction spots in the image. It can be just a constant 
    value, the diffracted beam energy (in keV) or the intensity as computed by the :py:meth:`diffracted_intensity` 
    method.

    :param orientation: The crystal orientation.
    :param detector: An instance of the Detector2d class.
    :param list hkl_planes: A list of the lattice planes to include in the pattern.
    :param Xu: The unit vector of the incident X-ray beam (default along the X-axis).
    :param bool use_friedel_pair: also consider the Friedel pair of each lattice plane in the list as candidate for diffraction. 
    :param spectrum: A two columns array of the spectrum to use for the calculation.
    :param float spectrum_thr: The threshold to use to determine if a wave length is contributing or not.
    :param int r_spot: Size of the spots on the detector in pixel (5 by default)
    :param str color_field: a traing describing, must be 'constant', 'energy' or 'intensity'
    :param bool inverted: A flag to control if the pattern needs to be inverted.
    :param bool show_direct_beam: A flag to control if the direct beam is shown.
    :param bool verbose: activate verbose mode (False by default).
    :return: the computed pattern as a numpy array.
    """
    detector.data = np.zeros(detector.size, dtype=np.float32)
    # create a small square image for one spot
    spot = np.ones((2 * r_spot + 1, 2 * r_spot + 1), dtype=np.uint8)
    max_val = np.iinfo(np.uint8).max  # 255 here
    direct_beam_lab = detector.project_along_direction(Xu)
    direct_beam_pix = detector.lab_to_pixel(direct_beam_lab)[0]
    if show_direct_beam:
        add_to_image(detector.data, max_val * 3 * spot, (direct_beam_pix[0], direct_beam_pix[1]))
    if spectrum is not None:
        print('using spectrum')
        #indices = np.argwhere(spectrum[:, 1] > spectrum_thr)
        E_min = min(spectrum) #float(spectrum[indices[0], 0])
        E_max = max(spectrum) #float(spectrum[indices[-1], 0])
        lambda_min = lambda_keV_to_nm(E_max)
        lambda_max = lambda_keV_to_nm(E_min)
        #if verbose:
        print('energy bounds: [{0:.1f}, {1:.1f}] keV'.format(E_min, E_max))

    for hkl in hkl_planes:
        (the_energy, theta) = select_lambda(hkl, orientation, Xu=Xu, verbose=False)
        if the_energy < 0:
            if use_friedel_pair:
                if verbose:
                    print('switching to Friedel pair')
                hkl = hkl.friedel_pair()
                (the_energy, theta) = select_lambda(hkl, orientation, Xu=Xu, verbose=False)
            else:
                continue
        assert(the_energy >= 0)
        if spectrum is not None:
            if the_energy < E_min or the_energy > E_max:
                #print('skipping reflection {0:s} which would diffract at {1:.1f}'.format(hkl.miller_indices(), abs(the_energy)))
                continue
                #print('including reflection {0:s} which will diffract at {1:.1f}'.format(hkl.miller_indices(), abs(the_energy)))
        K = diffracted_vector(hkl, orientation, Xu=Xu, use_friedel_pair=False, verbose=verbose)
        if K is None or np.dot(Xu, K) == 0:
            continue  # skip diffraction // to the detector
        d = np.dot((detector.ref_pos - np.array([0., 0., 0.])), detector.w_dir) / np.dot(K, detector.w_dir)
        if d < 0:
            if verbose:
                print('skipping diffraction not towards the detector')
            continue

        R = detector.project_along_direction(K, origin=[0., 0., 0.])
        (u, v) = detector.lab_to_pixel(R)[0]
        if verbose and u >= 0 and u < detector.size[0] and v >= 0 and v < detector.size[1]:
            print('* %d%d%d reflexion' % hkl.miller_indices())
            print('diffracted beam will hit the detector at (%.3f, %.3f) mm or (%d, %d) pixels' % (R[1], R[2], u, v))
            print('diffracted beam energy is {0:.1f} keV'.format(abs(the_energy)))
            print('Bragg angle is {0:.2f} deg'.format(abs(theta * 180 / pi)))
        # mark corresponding pixels on the image detector
        if color_field == 'constant':
            add_to_image(detector.data, max_val * spot, (u, v))
        elif color_field == 'energy':
            add_to_image(detector.data, abs(the_energy) * spot.astype(float), (u, v))
        elif color_field == 'intensity':
            I = diffracted_intensity(hkl, I0=max_val, verbose=verbose)
            add_to_image(detector.data, I * spot, (u, v))
        else:
            raise ValueError('unsupported color_field: %s' % color_field)
    if inverted:
        # np.invert works only with integer types
        print('casting to uint8 and inverting image')
        # limit maximum to max_val (255) and convert to uint8
        over = detector.data > 255
        detector.data[over] = 255
        detector.data = detector.data.astype(np.uint8)
        detector.data = np.invert(detector.data)
    return detector.data


def gnomonic_projection_point(data, OC=None):
    """compute the gnomonic projection of a given point or series of points in the general case.

    This methods *does not* assumes the incident X-ray beam is along (1, 0, 0). This is accounted for with the 
    parameter OC which indicates the center of the projection (the incident beam intersection with the detector). 
    A conditional treatment is done as the projection is is faster to compute in the case of normal incidence.

    The points coordinates are passed along with a single array which must be of size (n, 3) where n is the number of 
    points. If a single point is used, the data can indifferently be of size (1, 3) or (3). 

    :param ndarray data: array of the point(s) coordinates in the laboratory frame, aka OR components.
    :param ndarray OC: coordinates of the center of the gnomonic projection in the laboratory frame.
    :return data_gp: array of the projected point(s) coordinates in the laboratory frame.
    """
    if data.ndim == 1:
        data = data.reshape((1, 3))
    if OC is None:
        # fall back on normal incidence case
        r = np.sqrt(data[:, 1] ** 2 + data[:, 2] ** 2)  # mm
        theta = 0.5 * np.arctan(r / data[:, 0])
        p = data[:, 0] * np.tan(pi / 2 - theta)  # distance from the incident beam to the gnomonic projection mm
        data_gp = np.zeros_like(data)  # mm
        data_gp[:, 0] = data[:, 0]
        data_gp[:, 1] = - data[:, 1] * p / r
        data_gp[:, 2] = - data[:, 2] * p / r
    else:
        data_gp = np.zeros_like(data)  # mm
        data_gp[:, 0] = data[:, 0]
        for i in range(data.shape[0]):
            R = data[i]  # diffraction spot position in the laboratory frame
            CR = R - OC
            alpha = atan2(
                np.linalg.norm(np.cross(OC / np.linalg.norm(OC), CR / np.linalg.norm(CR))),
                np.dot(OC / np.linalg.norm(OC), CR / np.linalg.norm(CR))) - pi / 2
            # the Bragg angle can be measured using the diffraction spot position
            theta = 0.5 * np.arccos(np.dot(OC / np.linalg.norm(OC), R / np.linalg.norm(R)))
            r = np.sqrt(CR[1] ** 2 + CR[2] ** 2)  # mm
            # distance from the incident beam to the gnomonic projection mm
            p = np.linalg.norm(OC) * cos(theta) / sin(theta - alpha)
            data_gp[i, 1] = OC[1] - CR[1] * p / r
            data_gp[i, 2] = OC[2] - CR[2] * p / r
    return data_gp


def gnomonic_projection(detector, pixel_size=None, OC=None, verbose=False):
    """This function carries out the gnomonic projection of the detector image.
    
    The data must be of uint8 type (between 0 and 255) with diffraction spots equals to 255.
    The function create a new detector instance (think of it as a virtual detector) located at the same position 
    as the given detector and with an inverse pixel size. The gnomonic projection is stored into this new detector data.
    The gnomonic projection of each white pixel (value at 255) is computed. The projection is carried out with respect 
    to the center detector (ucen, vcen) point.
    
    :param RegArrayDetector2d detector: the detector instance with the data from which to compute the projection.
    :param float pixel_size: pixel size to use in the virtual detector for the gnomonic projection.
    :param tuple OC: coordinates of the center of the gnomonic projection in the laboratory frame.
    :param bool verbose: flag to activate verbose mode.
    :returns RegArrayDetector2d gnom: A virtual detector with the gnomonic projection as its data.
    """
    assert detector.data.dtype == np.uint8
    dif = detector.data == 255  # boolean array used to select pixels with diffracted intensity
    dif_indices = np.where(dif)  # (ui, vi) tuple with 1D arrays of the coordinates u and v
    n = dif_indices[0].shape
    if verbose:
        print('%d points in the gnomonic projection' % n)
        print('center of the projection: %s' % OC)
    uv_mm = detector.pixel_to_lab(dif_indices[0], dif_indices[1])
    uvg_mm = gnomonic_projection_point(uv_mm, OC)  # mm
    print(uvg_mm.shape)

    # create the new virtual detector
    from pymicro.xray.detectors import RegArrayDetector2d
    gnom = RegArrayDetector2d(size=np.array(detector.size))
    gnom.ref_pos = detector.ref_pos  # same ref position as the actual detector
    if not pixel_size:
        gnom.pixel_size = 1. / detector.pixel_size  # mm
    else:
        gnom.pixel_size = pixel_size  # mm
    # TODO remove the next two lines
    gnom.ucen = detector.ucen
    gnom.vcen = detector.vcen

    # create the gnom.data array (zeros with pixels set to 1 for gnomonic projection points)
    gnom.data = np.zeros(gnom.size, dtype=np.uint8)
    # uvg_px = gnom.lab_to_pixel(uvg_mm)
    uvg_px = np.zeros((uvg_mm.shape[0], 2), dtype=np.int)
    for i in range(uvg_mm.shape[0]):
        uvg_px[i, :] = gnom.lab_to_pixel(uvg_mm[i, :])
    # filter out point outside the virtual detector
    detin = (uvg_px[:, 0] >= 0) & (uvg_px[:, 0] <= gnom.size[0] - 1) & \
            (uvg_px[:, 1] >= 0) & (uvg_px[:, 1] <= gnom.size[1] - 1)
    gnom.data[uvg_px[detin, 0], uvg_px[detin, 1]] = 1
    return gnom


def identify_hkl_from_list(hkl_list):
    if hkl_list[1] == hkl_list[2]:
        ids = [1, 3, 0]
    elif hkl_list[1] == hkl_list[3]:
        ids = [1, 2, 0]
    elif hkl_list[0] == hkl_list[2]:
        ids = [0, 3, 1]
    elif hkl_list[0] == hkl_list[3]:
        ids = [0, 2, 1]
    '''
    if hkl_list[1] == hkl_list[2]:
        if hkl_list[3] == hkl_list[4] and hkl_list[0] == hkl_list[5] or \
                                hkl_list[3] == hkl_list[5] and hkl_list[0] == hkl_list[4]:
            identified_list = [hkl_list[1], hkl_list[3], hkl_list[0]]

    elif hkl_list[1] == hkl_list[3]:
        if hkl_list[2] == hkl_list[4]:
            if hkl_list[0] == hkl_list[5]:
                identified_list = [hkl_list[1], hkl_list[2], hkl_list[0]]

        elif hkl_list[2] == hkl_list[5]:
            if hkl_list[0] == hkl_list[4]:
                identified_list = [hkl_list[1], hkl_list[2], hkl_list[0]]

    elif hkl_list[0] == hkl_list[2]:
        if hkl_list[3] == hkl_list[4]:
            if hkl_list[1] == hkl_list[5]:
                identified_list = [hkl_list[0], hkl_list[3], hkl_list[1]]

        elif hkl_list[3] == hkl_list[5]:
            if hkl_list[1] == hkl_list[4]:
                identified_list = [hkl_list[0], hkl_list[3], hkl_list[1]]

    elif hkl_list[0] == hkl_list[3]:
        if hkl_list[2] == hkl_list[4]:
            if hkl_list[1] == hkl_list[5]:
                identified_list = [hkl_list[0], hkl_list[2], hkl_list[1]]

        elif hkl_list[2] == hkl_list[5]:
            if hkl_list[1] == hkl_list[4]:
                identified_list = [hkl_list[0], hkl_list[2], hkl_list[1]]
    '''
    return [hkl_list[i] for i in ids]


def triplet_indexing(OP, angles_exp, angles_th, tol=1.0, verbose=False):
    '''
    evaluate each triplet composed by 3 diffracted points.
    the total number of triplet is given by the binomial coefficient (n, 3) = n*(n-1)*(n-2)/6
    '''
    nombre_triplet = 0
    OP_indexed = []
    for ti in range(len(OP) - 2):

        for tj in range(ti + 1, len(OP) - 1):

            triplet = np.zeros(3)
            for tk in range(tj + 1, len(OP)):
                nombre_triplet = nombre_triplet + 1
                print('\n\ntriplet number = {0:d}'.format(nombre_triplet))
                # look for possible matches for a given triplet
                triplet = [(ti, tj), (tj, tk), (ti, tk)]
                print('** testing triplet %s' % triplet)
                candidates = []  # will contain the 3 lists of candidates for a given triplet
                for (i, j) in triplet:
                    angle_to_match = angles_exp[i, j]
                    # find the index by looking to matching angles within the tolerance
                    cand = np.argwhere(
                        abs(angles_th['angle'] - angle_to_match) < tol).flatten()  # flatten since shape is (n, 1)
                    print('* couple (OP%d, OP%d) has angle %.2f -> %d couple candidate(s) in the angle_th list: %s' % (
                        i, j, angle_to_match, len(cand), cand))
                    candidates.append(cand)
                print('looking for real triplet in all %d candidates pairs:' % (
                    len(candidates[0]) * len(candidates[1]) * len(candidates[2])))
                for ci in candidates[0]:
                    for cj in candidates[1]:
                        for ck in candidates[2]:
                            # make a list of the candidate hkl plane indices, based on the angle match
                            hkl_list = [angles_th['hkl1'][ci], angles_th['hkl2'][ci],
                                        angles_th['hkl1'][cj], angles_th['hkl2'][cj],
                                        angles_th['hkl1'][ck], angles_th['hkl2'][ck]]
                            # look how many unique planes are in the list (should be 3 for a triplet)
                            if verbose:
                                print('- candidates with indices %s correspond to %s' % ([ci, cj, ck], hkl_list))
                            unique_indices = np.unique(hkl_list)
                            if len(unique_indices) != 3:
                                if verbose:
                                    print('not a real triplet, skipping this one')
                                continue
                            print('this is a triplet: %s' % unique_indices)
                            # the points can now be identified from the hkl_list ordering
                            identified_list = identify_hkl_from_list(hkl_list)
                            full_list = [tj, tk, ti]
                            full_list.extend(identified_list)
                            print(full_list)
                            OP_indexed.append(full_list)
                            print('*  (OP%d) -> %d ' % (tj, identified_list[0]))
                            print('*  (OP%d) -> %d ' % (tk, identified_list[1]))
                            print('*  (OP%d) -> %d ' % (ti, identified_list[2]))
    print('indexed list length is %d' % len(OP_indexed))
    print(OP_indexed)
    return OP_indexed


def transformation_matrix(hkl_plane_1, hkl_plane_2, xyz_normal_1, xyz_normal_2):
    # create the vectors representing this frame in the crystal coordinate system
    e1_hat_c = hkl_plane_1.normal()
    e2_hat_c = np.cross(hkl_plane_1.normal(), hkl_plane_2.normal()) / np.linalg.norm(
        np.cross(hkl_plane_1.normal(), hkl_plane_2.normal()))
    e3_hat_c = np.cross(e1_hat_c, e2_hat_c)
    e_hat_c = np.array([e1_hat_c, e2_hat_c, e3_hat_c])
    # create local frame attached to the indexed crystallographic features in XYZ
    e1_hat_star = xyz_normal_1
    e2_hat_star = np.cross(xyz_normal_1, xyz_normal_2) / np.linalg.norm(
        np.cross(xyz_normal_1, xyz_normal_2))
    e3_hat_star = np.cross(e1_hat_star, e2_hat_star)
    e_hat_star = np.array([e1_hat_star, e2_hat_star, e3_hat_star])
    # now build the orientation matrix
    orientation_matrix = np.dot(e_hat_c.T, e_hat_star)
    return orientation_matrix


def confidence_index(votes):
    n = np.sum(votes)  # total number of votes
    v1 = max(votes)
    votes[votes.index(v1)] = 0.
    v2 = max(votes)
    ci = (1. * (v1 - v2)) / n
    return ci


def poll_system(g_list, dis_tol=1.0, verbose=False):
    """
    Poll system to sort a series of orientation matrices determined by the indexation procedure.

    For each orientation matrix, check if it corresponds to an existing solution, if so: vote for it,
    if not add a new solution to the list
    :param list g_list: the list of orientation matrices (should be in the fz)
    :param float dis_tol: angular tolerance (degrees)
    :param bool verbose: activate verbose mode (False by default)
    :return: a tuple composed by the most popular orientation matrix, the corresponding vote number and the confidence index
    """
    solution_indices = [0]
    votes = [0]
    vote_index = np.zeros(len(g_list), dtype=int)
    dis_tol_rad = dis_tol * pi / 180
    from pymicro.crystal.microstructure import Orientation
    for i in range(len(g_list)):
        g = g_list[i]
        # rotations are already in the fundamental zone
        for j in range(len(solution_indices)):
            index = solution_indices[j]
            delta = np.dot(g, g_list[index].T)
            # compute misorientation angle in radians
            angle = Orientation.misorientation_angle_from_delta(delta)
            if verbose:
                print('j=%d -- angle=%f' % (j, angle))
            if angle <= dis_tol_rad:
                votes[j] += 1
                vote_index[i] = j
                if verbose:
                    print('angle (deg) is %.2f' % (180 / pi * angle))
                    print('vote list is now %s' % votes)
                    print('solution_indices list is now %s' % solution_indices)
                break
            elif j == len(solution_indices) - 1:
                solution_indices.append(i)
                votes.append(1)
                vote_index[i] = len(votes) - 1
                if verbose:
                    print('vote list is now %s' % votes)
                    print('solution_indices list is now %s' % solution_indices)
                break
    index_result = np.argwhere(votes == np.amax(votes)).flatten()
    if verbose:
        print('Max vote =', np.amax(votes))
        print('index result:', index_result)
        print('Number of equivalent solutions :', len(index_result))
        print(type(index_result))
        print(index_result.shape)
    final_orientation_matrix = []
    for n in range(len(index_result)):
        solutions = g_list[solution_indices[index_result[n]]]
        if verbose:
            print('Solution number {0:d} is'.format(n+1), solutions)
        final_orientation_matrix.append(solutions)
    result_vote = max(votes)
    ci = confidence_index(votes)
    vote_field = [votes[i] for i in vote_index]
    return final_orientation_matrix, result_vote, ci, vote_field


def index(hkl_normals, hkl_planes, tol_angle=0.5, tol_disorientation=1.0, symmetry=Symmetry.cubic, display=False):
    # angles between normal from the gnomonic projection
    angles_exp = np.zeros((len(hkl_normals), len(hkl_normals)), dtype=float)
    print('\nlist of angles between points on the detector')
    for i in range(len(hkl_normals)):
        for j in range(i + 1, len(hkl_normals)):
            angle = 180 / pi * np.arccos(np.dot(hkl_normals[i], hkl_normals[j]))
            angles_exp[i, j] = angle
            print('%.2f, OP%d, OP%d' % (angles_exp[i, j], i, j))
    # keep a list of the hkl values as string
    hkl_str = []
    for p in hkl_planes:
        (h, k, l) = p.miller_indices()
        hkl_str.append('(%d%d%d)' % (h, k, l))
    # compute theoretical angle between each plane normal, store the results using a structured array
    angles_th = np.empty(len(hkl_planes) * (len(hkl_planes) - 1) // 2,
                         dtype=[('angle', 'f4'), ('hkl1', 'i4'), ('hkl2', 'i4')])
    index = 0
    for i in range(len(hkl_planes)):
        for j in range(i + 1, len(hkl_planes)):
            angle = 180 / pi * np.arccos(np.dot(hkl_planes[i].normal(), hkl_planes[j].normal()))
            angles_th[index] = (angle, i, j)
            index += 1

    # sort the array by increasing angle
    angles_th.sort(order='angle')
    print('\nsorted list of angles between hkl plane normals')
    for i in range(len(angles_th)):
        print('%.3f, (%d, %d) -> %s, %s' % (
            angles_th['angle'][i], angles_th['hkl1'][i], angles_th['hkl2'][i], hkl_str[angles_th['hkl1'][i]],
            hkl_str[angles_th['hkl2'][i]]))

    # index by triplets
    normal_indexed = triplet_indexing(hkl_normals, angles_exp, angles_th, tol=tol_angle)
    print('indexed list length is %d' % len(normal_indexed))
    print(normal_indexed)

    # Compute the orientation matrix g for all different triplets
    g_indexation = []
    for i in range(len(normal_indexed)):
        # a given indexed triplet allow to construct 3 orientation matrices (which should be identical)
        pos = [[0, 1, 3, 4], [1, 2, 4, 5], [2, 0, 5, 3]]
        for j in range(3):
            orientation_matrix = transformation_matrix(
                hkl_planes[normal_indexed[i][pos[j][2]]], hkl_planes[normal_indexed[i][pos[j][3]]],
                hkl_normals[normal_indexed[i][pos[j][0]]], hkl_normals[normal_indexed[i][pos[j][1]]])
        # move to the fundamental zone
        om_fz = symmetry.move_rotation_to_FZ(orientation_matrix)  # we only add the third one
        g_indexation.append(om_fz)
    final_orientation_matrix, vote, ci, vote_field = poll_system(g_indexation, dis_tol=tol_disorientation)
    if final_orientation_matrix == 0:
        print('Troubles in the data set !')
    else:
        print('\n\n\n### FINAL SOLUTION(S) ###\n')
        for n in range(len(final_orientation_matrix)):
            print('- SOLUTION %d -' % (n + 1))
            final_orientation = Orientation(final_orientation_matrix[n])
            print(final_orientation.inFZ())
            print('- Cristal orientation in Fundamental Zone \n {} \n'.format(final_orientation.euler))
            print('- Rodrigues vector in the fundamental Zone \n {} \n'.format(final_orientation.rod))
            if display:
                from pymicro.crystal.texture import PoleFigure
                PoleFigure.plot(final_orientation, axis='Z')
        return final_orientation_matrix, ci

def zone_axis_list(angle, orientation, lattice,  max_miller=5,  Xu=np.array([1., 0., 0.]), verbose=False):
    """
    This function allows to get easily the Miller indices of zone axis present in a pattern.

    :param float list angle: the angle max or the zone axis angle range admissible around the detector center.
    :param orientation: The orientation of the crystal lattice.
    :param lattice: The corresponding crystal lattice, instance of Lattice.
    :param int max_miller: Maximal value allowed of Miller indices direction.
    :param array Xu: The unit vector of the incident X-ray beam (default along the X-axis).
    :param bool verbose: activate verbose mode (default False).
    :return: A list of HklDirection instance of all zone axis in the angle range.
    """
    ZA_list = []
    if len(angle) == 1:
        angle_max = angle[0] * np.pi / 180.
        angle_min = 0.
    if len(angle) == 2:
        angle_max = max(angle) * np.pi / 180.
        angle_min = min(angle) * np.pi / 180.
        print("Get the indices of directions between [%d, %d] degrees" %(min(angle), max(angle)))
    indices = range(-max_miller, max_miller+1)
    for h in indices:
        for k in indices:
            for l in indices:
                if h == k == l == 0:  # skip (0, 0, 0)
                    continue
                uvw = HklDirection(h, k, l, lattice)
                # compute the direction in the lab frame
                ZA = np.dot(orientation.orientation_matrix().transpose(), uvw.direction())
                psi = np.arccos(np.dot(ZA, Xu))
                if angle_min < psi < angle_max:
                    if verbose == True:
                        print('found zone axis [%d%d%d] at %.1f deg from incident beam' % (h,k,l,(psi*180/pi)))
                    ZA_list.append(uvw) #zones axis list which are inferior with max angle
    ZA_list = HklObject.skip_higher_order(ZA_list)
    return ZA_list

def get_gnomonic_edges(detector, gnom, OC=None, num_points=21):
    """
    This function allows to get the blind area of the gnomonic projection.

    :param RegArrayDetector2d detector: the detector instance with the data from which to compute the projection.
    :param RegArrayDetector2d gnom: A virtual detector with the gnomonic projection as its data.
    :param ndarray OC: coordinates of the center of the gnomonic projection in the laboratory frame.
    :param int num_points: number of points to describe an edge (minimum 2)
    :return: ndarray of gnomonic blind area edges coordinates.
    """
    # Get detector pixel edges
    uv_detector_edges = detector.get_edges(num_points=num_points, verbose=False)  # pixels
    # Compute edges position in the lab coordinates
    detector_edges_mm = detector.pixel_to_lab(uv_detector_edges[:, 0], uv_detector_edges[:, 1])
    # Apply the gnomonic projection  in the lab coordinates
    detector_edges_gp = gnomonic_projection_point(detector_edges_mm, OC=OC)  # mm
    # return the gnomonic projected point on the detector (pixel coordinates)
    return gnom.lab_to_pixel(detector_edges_gp)

def diffracting_normals_vector(gnom):
    """
    This function allows to get easily the diffracting normal vector from the gnomonic projection images.

    :param RegArrayDetector2d gnom: A virtual detector with the gnomonic projection as its data.
    :return: Normalized normal vector of diffracting plane
    """
    uv_g = np.argwhere(gnom.data == 1)  # points on the gnomonic projection
    OP = [gnom.pixel_to_lab(uv_g[i, 0], uv_g[i, 1]).tolist() for i in range(len(uv_g))]
    hkl_normals = [(n / np.linalg.norm(n)).tolist() for n in OP]  # normalized list of vectors
    print('%d normals found in the gnomonic projection' % len(hkl_normals))

    return hkl_normals

from pymicro.xray.experiment import ForwardSimulation, Experiment

class LaueForwardSimulation(ForwardSimulation):
    """Class to represent a Forward Simulation."""

    def __init__(self, verbose=False):
        super(LaueForwardSimulation, self).__init__('laue', verbose=verbose)
        self.hkl_planes = []
        self.max_miller = 5
        self.use_energy_limits = False
        self.exp = Experiment()

    def set_experiment(self, experiment):
        """Attach an X-ray experiment to this simulation."""
        self.exp = experiment

    def set_use_energy_limits(self, use_energy_limits):
        """Activate or deactivate the use of energy limits."""
        self.use_energy_limits = use_energy_limits

    def set_hkl_planes(self, hkl_planes):
        self.hkl_planes = hkl_planes

    def setup(self, include_grains=None):
        """Setup the forward simulation."""
        pass

    @staticmethod
    def fsim_laue(orientation, hkl_planes, positions, source_position):
        """Simulate Laue diffraction conditions based on a crystal orientation, a set of lattice planes and physical
        positions.

        This function is the work horse of the forward model. It uses a set of HklPlane instances and the voxel
        coordinates to compute the diffraction quantities.

        :param Orientation orientation: the crystal orientation.
        :param list hkl_planes: a list of `HklPlane` instances.
        :param list positions: a list of (x, y, z) positions.
        :param tuple source_position: a (x, y, z) tuple describing the source position.
        """
        n_hkl = len(hkl_planes)
        n_vox = len(positions)
        gt = orientation.orientation_matrix().transpose()

        # here we use list comprehension to avoid for loops
        d_spacings = [hkl.interplanar_spacing() for hkl in hkl_planes]  # size n_hkl
        G_vectors = [hkl.scattering_vector() for hkl in hkl_planes]  # size n_hkl, with 3 elements items
        Gs_vectors = [gt.dot(Gc) for Gc in G_vectors]  # size n_hkl, with 3 elements items
        Xu_vectors = [(pos - source_position) / np.linalg.norm(pos - source_position)
                      for pos in positions]  # size n_vox
        thetas = [np.arccos(np.dot(Xu, Gs / np.linalg.norm(Gs))) - np.pi / 2
                  for Xu in Xu_vectors
                  for Gs in Gs_vectors]  # size n_vox * n_hkl
        the_energies = [lambda_nm_to_keV(2 * d_spacings[i_hkl] * np.sin(thetas[i_Xu * n_hkl + i_hkl]))
                        for i_Xu in range(n_vox)
                        for i_hkl in range(n_hkl)]  # size n_vox * n_hkl
        X_vectors = [np.array(Xu_vectors[i_Xu]) / 1.2398 * the_energies[i_Xu * n_hkl + i_hkl]
                     for i_Xu in range(n_vox)
                     for i_hkl in range(n_hkl)]  # size n_vox * n_hkl
        K_vectors = [X_vectors[i_Xu * n_hkl + i_hkl] + Gs_vectors[i_hkl]
                     for i_Xu in range(n_vox)
                     for i_hkl in range(n_hkl)]  # size n_vox * n_hkl
        return Xu_vectors, thetas, the_energies, X_vectors, K_vectors

    def fsim_grain(self, gid=1):
        self.grain = self.exp.get_sample().get_microstructure().get_grain(gid)
        sample = self.exp.get_sample()
        lattice = sample.get_microstructure().get_lattice()
        source = self.exp.get_source()
        detector = self.exp.get_active_detector()
        data = np.zeros_like(detector.data)
        if self.verbose:
            print('Forward Simulation for grain %d' % self.grain.id)
        sample.geo.discretize_geometry(grain_id=self.grain.id)
        # we use either the hkl planes for this grain or the ones defined for the whole simulation
        if hasattr(self.grain, 'hkl_planes') and len(self.grain.hkl_planes) > 0:
            print('using hkl from the grain')
            hkl_planes = [HklPlane(h, k, l, lattice) for (h, k, l) in self.grain.hkl_planes]
        else:
            if len(self.hkl_planes) == 0:
                print('warning: no reflection defined for this simulation, using all planes with max miller=%d' % self.max_miller)
                self.set_hkl_planes(build_list(lattice=lattice, max_miller=self.max_miller))
            hkl_planes = self.hkl_planes
        n_hkl = len(hkl_planes)
        positions = sample.geo.get_positions()  # size n_vox, with 3 elements items
        n_vox = len(positions)
        Xu_vectors, thetas, the_energies, X_vectors, K_vectors = LaueForwardSimulation.fsim_laue(
            self.grain.orientation, hkl_planes, positions, source.position)
        OR_vectors = [detector.project_along_direction(origin=positions[i_vox],
                                                       direction=K_vectors[i_vox * n_hkl + i_hkl])
                      for i_vox in range(n_vox)
                      for i_hkl in range(n_hkl)]  # size nb_vox * n_hkl
        uv = [detector.lab_to_pixel(OR)[0].astype(np.int)
              for OR in OR_vectors]
        # now construct a boolean list to select the diffraction spots
        if source.min_energy is None and source.max_energy is None:
            # TODO use the use_energy_limits attribute
            energy_in = [True for k in range(len(the_energies))]
        else:
            energy_in = [source.min_energy < the_energies[k] < source.max_energy
                         for k in range(len(the_energies))]
        uv_in = [0 < uv[k][0] < detector.get_size_px()[0] and 0 < uv[k][1] < detector.get_size_px()[1]
                 for k in range(len(uv))]  # size n, diffraction located on the detector
        spot_in = [uv_in[k] and energy_in[k] for k in range(len(uv))]
        if self.verbose:
            print('%d diffraction events on the detector among %d' % (sum(spot_in), len(uv)))

        # now sum the counts on the detector individual pixels
        for k in range(len(uv)):
            if spot_in[k]:
                data[uv[k][0], uv[k][1]] += 1
        return data

    def fsim(self):
        """run the forward simulation.

        If the sample has a CAD type of geometry, a single grain (the first from the list) is assumed. In the other
        cases all the grains from the microstructure are used. In particular, if the microstructure has a grain map,
        it can be used to carry out an extended sample simulation.
        """
        full_data = np.zeros_like(self.exp.get_active_detector().data)
        micro = self.exp.get_sample().get_microstructure()
        # for cad geometry we assume only one grain (the first in the list)
        if self.exp.get_sample().geo.geo_type == 'cad':
            full_data += self.fsim_grain(gid=micro.grains[0].id)
        else:
            # in the other cases, we use all the grains defined in the microstructure
            for grain in micro.grains:
                full_data += self.fsim_grain(gid=grain.id)
        return full_data




import numpy as np

class Quaternion:
    """Class to describe a Quaternion."""

    def __init__(self, array, convention=1):
        assert len(array) == 4
        array = np.array(array)
        self.quat = array / np.sqrt(np.sum(array ** 2))
        self.convention = convention  # active by default

    def q0(self):
        return self.quat[0]

    def q1(self):
        return self.quat[1]

    def q2(self):
        return self.quat[2]

    def q3(self):
        return self.quat[3]

    def __repr__(self):
        return str(self.quat)

    def norm(self):
        """Compute the norm of the quaternion (should be 1)."""
        return np.sqrt(np.sum(self.quat ** 2))



import os, sys
import numpy as np
import struct


def read_image_sequence(data_dir, prefix, num_images, start_index=0, image_format='png', zero_padding=0, crop=None, verbose=False):
    '''Read a series of images into a list of numpy arrays.
    
    :param str data_dir: directory where the image files are located.
    :param str prefix: a string to construct the image file names.
    :param int num_images: the number of images to read.
    :param int start_index: the index to start loading the images (0 by default).
    :param str image_format: can be tif or png (png by default).
    :param int zero_padding: number of zero to use in zero padding (0 by default).
    :param list crop: bounds to crop the images (None by default) 
    :param bool verbose: activate verbose mode (False by default).
    :return: the list of the images read from the disk.
    '''
    # build the numbering pattern
    pat = '0%dd' % zero_padding
    fmt = '{0:s}{1:' + pat + '}.{2:s}'
    image_stack = []
    if image_format == 'tif':
        from pymicro.external.tifffile import TiffFile
    elif image_format == 'png':
        from matplotlib import pyplot as plt
    for i in range(start_index, start_index + num_images):
        image_path = os.path.join(data_dir, fmt.format(prefix, i, image_format))
        if verbose:
            print('loading image %s' % image_path)
        if image_format == 'tif':
            im = TiffFile(image_path).asarray()
        elif image_format == 'png':
            im = (np.mean(plt.imread(image_path)[:, :, :3], axis=2) * 255).astype(np.uint8)
        if crop:
            im = im[crop[2]:crop[3], crop[0]:crop[1]]
        image_stack.append(im.T)
    return image_stack


def unpack_header(header):
    """Unpack an ascii header.

    Form a string with the read binary data and then split it into string
    tokens which are put in a dictionnary.

    :param str header: the ascii header to unpack.
    :return: a dictionnary with the (key, value) fields contained in the header.
    """
    header_values = {}
    for line in header.split('\n'):
        tokens = line.split('=')
        if len(tokens) > 1:
            header_values[tokens[0].strip()] = tokens[1].split(';')[0].strip()
    return header_values


def edf_info(file_name, header_size=None, verbose=False):
    """Read and return informations contained in the header of a .edf file.

    Edf files always start with a header (of variable length) containing
    informations about the file such as acquisition conditions, image
    dimensions... This function reads a certain amount of bytes of a given
    file as ascii data and unpack it.
    If not specified, the header size is determined automatically by
    substracting the data size (read as ascii at the begining of the file)
    to the total file size.

    :param str file_name: the name of the edf file to read.
    :param int header_size: number of bytes to read as a multiple of 512 (None by default).
    :param bool verbose: flag to activate verbose mode.
    :return: a dictionnary containing the file informations.
    """
    try:
        f = open(file_name, 'r', encoding='latin-1')
    except TypeError:
        f = open(file_name, 'r')  # fall back
    if header_size is None:
        # guess the header size by peeking at the first chunk of 512 bytes
        header_values = unpack_header(f.read(512))
        total_file_size = os.path.getsize(file_name)
        payload_size = int(header_values['Size'].split('.')[0])
        header_size = total_file_size - payload_size
        if verbose:
            print('determined header size is %d bytes' % header_size)
        f.seek(0)
    header = f.read(header_size)
    f.close()
    return unpack_header(header)


def edf_read(file_name, verbose=False):
    """Read an edf file.

    edf stands for ESRF data file. It has a variable header size which is
    a multiple of 512 bytes and contains the image meta in ASCII format
    (eg. image size, data type, motor positions).

    The ascii header is parsed automatically by `edf_info` to retreive the
    image size and data type. Depending on the information enclosed in the
    header, this function may return a 1d, 2d or 3d array.
    ::

      >>> im = edf_read('radio.edf')
      >>> im.shape
      (2048, 2048)

    :param str file_name: the name of the edf file to read.
    :param bool verbose: flag to activate verbose mode.
    :return: a numpy array containing the data
    """
    header_values = edf_info(file_name, verbose=verbose)
    f = open(file_name, 'r')
    data_type = esrf_to_numpy_datatype(header_values['DataType'])
    if verbose:
        print(header_values['DataType'], data_type)
    # get the payload size
    payload_size = int(header_values['Size'].split('.')[0])
    # get the image size from the ascii header
    dim_1 = int(header_values['Dim_1'].split('.')[0])
    try:
        dim_2 = int(header_values['Dim_2'].split('.')[0])
    except KeyError:
        if verbose:
            print('Dim_2 not defined in header')
        dim_2 = None
    try:
        dim_3 = int(header_values['Dim_3'].split('.')[0])
    except KeyError:
        if verbose:
            print('Dim_3 not defined in header')
        dim_3 = None
    # now read binary data
    header_size = os.path.getsize(file_name) - payload_size
    f.seek(header_size)
    payload = np.fromfile(f, dtype=data_type)
    if dim_1 and dim_2 and dim_3:
        data = np.reshape(payload, (dim_3, dim_2, dim_1)).transpose(2, 1, 0)
    elif dim_1 and dim_2:
        data = np.reshape(payload, (dim_2, dim_1)).transpose(1, 0)
    else:
        data = np.reshape(payload, (dim_1))
    f.close()
    # pay attention to byte order
    if header_values['ByteOrder'] == 'HighByteFirst':
        data = data.byteswap()
    return data


def esrf_to_numpy_datatype(data_type):
    return {
        'UnsignedByte': np.uint8,
        'UnsignedShort': np.uint16,
        'UnsignedLong': np.uint32,
        'SignedInteger': np.int32,
        'FloatValue': np.float32,
        'DoubleValue': np.float64,
    }.get(data_type, np.uint16)


def numpy_to_esrf_datatype(data_type):
    return {
        np.uint8: 'UnsignedByte',
        np.uint16: 'UnsignedShort',
        np.uint32: 'UnsignedLong',
        np.int32: 'SignedInteger',
        np.float32: 'FloatValue',
        np.float64: 'DoubleValue',
        np.dtype('uint8'): 'UnsignedByte',
        np.dtype('uint16'): 'UnsignedShort',
        np.dtype('uint32'): 'UnsignedLong',
        np.dtype('int32'): 'SignedInteger',
        np.dtype('float32'): 'FloatValue',
        np.dtype('float64'): 'DoubleValue',
        float: 'DoubleValue',
    }.get(data_type, 'UnsignedShort')


def edf_write(data, file_name, header_size=1024):
    """Write a binary edf file with the appropriate header.

    This function write a (x,y,z) 3D dataset to the disk.
    The file is written as a Z-stack. It means that the first nx*ny bytes
    represent the first slice and so on...

    :param ndarray data: the data array to write to the file.
    :param str file_name: the file name to use.
    :param int header_size: the size of te header (a multiple of 512).
    """
    # get current time
    from time import gmtime, strftime
    today = strftime('%d-%b-%Y', gmtime())
    size = np.shape(data)
    print('data size in pixels is ', size)
    nbytes = np.prod(size) * data.dtype.itemsize
    print('opening', file_name, 'for writing')
    # craft an ascii header of the appropriate size
    f = open(file_name, 'wb')
    head = '{\n'
    head += 'HeaderID       = EH:000001:000000:000000 ;\n'
    head += 'Image          = 1 ;\n'
    head += 'ByteOrder      = LowByteFirst ;\n'
    head += 'DataType       = %13s;\n' % numpy_to_esrf_datatype(data.dtype)
    print('using data type %s' % numpy_to_esrf_datatype(data.dtype))
    head += 'Dim_1          = %4s;\n' % size[0]
    if len(size) > 1: head += 'Dim_2          = %4s;\n' % size[1]
    if len(size) > 2: head += 'Dim_3          = %4s;\n' % size[2]
    head += 'Size           = %9s;\n' % nbytes
    head += 'Date           = ' + today + ' ;\n'
    for i in range(header_size - len(head) - 2):
        head += ' '
    head += '}\n'
    f.write(head.encode('utf-8'))
    if len(data.shape) == 3:
        s = np.ravel(data.transpose(2, 1, 0)).tostring()
    elif len(data.shape) == 2:
        s = np.ravel(data.transpose(1, 0)).tostring()
    else:
        s = np.ravel(data).tostring()
    f.write(s)
    f.close()


def HST_info(info_file):
    """Read the given info file and returns a dictionary containing the data size and type.
    
    .. note::
    
       The first line of the file must begin by ! PyHST or directly by NUM_X. 
       Also note that if the data type is not specified, it will not be present in the dictionary.
    
    :param str info_file: path to the ascii file to read.
    :return: a dictionary with the values for x_dim, y_dim, z_dim and data_type if needed.
    """
    info_values = {}
    f = open(info_file, 'r')
    # the first line must contain PyHST or NUM_X
    line = f.readline()
    if line.startswith('! PyHST'):
        # read an extra line
        line = f.readline()
    elif line.startswith('NUM_X'):
        pass
    else:
        sys.exit('The file does not seem to be a PyHST info file')
    info_values['x_dim'] = int(line.split()[2])
    info_values['y_dim'] = int(f.readline().split()[2])
    info_values['z_dim'] = int(f.readline().split()[2])
    try:
        info_values['data_type'] = f.readline().split()[2]
    except IndexError:
        pass
    return info_values


def HST_read(scan_name, zrange=None, data_type=np.uint8, verbose=False,
             header_size=0, autoparse_filename=False, dims=None, mmap=False, pack_binary=False):
    '''Read a volume file stored as a concatenated stack of binary images.

    The volume size must be specified by dims=(nx, ny, nz) unless an associated
    .info file is present in the same location to determine the volume
    size. The data type is unsigned short (8 bits) by default but can be set
    to any numpy type (32 bits float for example).

    The autoparse_filename can be activated to retreive image type and
    size:
    ::

      HST_read(myvol_100x200x50_uint16.raw, autoparse_filename=True)

    will read the 3d image as unsigned 16 bits with size 100 x 200 x 50.

    .. note::

      If you use this function to read a .edf file written by
      matlab in +y+x+z convention (column major order), you may want to
      use: np.swapaxes(HST_read('file.edf', ...), 0, 1)

    :param str scan_name: path to the binary file to read.
    :param zrange: range of slices to use.
    :param data_type: numpy data type to use.
    :param bool verbose: flag to activate verbose mode.
    :param int header_size: number of bytes to skeep before reading the payload.
    :param bool autoparse_filename: flag to parse the file name to retreive the dims and data_type automatically.
    :param tuple dims: a tuple containing the array dimensions.
    :param bool mmap: activate the memory mapping mode.
    :param bool pack_binary: this flag should be true when reading a file written with the binary packing mode.
    '''
    if autoparse_filename:
        s_type = scan_name[:-4].split('_')[-1]
        data_type = np.dtype(s_type)
        s_size = scan_name[:-4].split('_')[-2].split('x')
        dims = (int(s_size[0]), int(s_size[1]), int(s_size[2]))
        if verbose:
            print('auto parsing filename: data type is set to', data_type)
    if dims is None:
        infos = HST_info(scan_name + '.info')
        [nx, ny, nz] = [infos['x_dim'], infos['y_dim'], infos['z_dim']]
        if 'data_type' in infos:
            if infos['data_type'] == 'PACKED_BINARY':
                pack_binary = True
                data_type = np.uint8
            else:
                data_type = np.dtype(infos['data_type'].lower())  # overwrite defaults with .info file value
    else:
        (nx, ny, nz) = dims
    if zrange is None:
        zrange = range(0, nz)
    if verbose:
        print('data type is', data_type)
        print('volume size is %d x %d x %d' % (nx, ny, len(zrange)))
        if pack_binary:
            print('unpacking binary data from single bytes (8 values per byte)')
    if mmap:
        data = np.memmap(scan_name, dtype=data_type, mode='c', shape=(len(zrange), ny, nx))
    else:
        f = open(scan_name, 'rb')
        f.seek(header_size + np.dtype(data_type).itemsize * nx * ny * zrange[0])
        if verbose:
            print('reading volume... from byte %d' % f.tell())
        # read the payload
        payload = f.read(np.dtype(data_type).itemsize * len(zrange) * ny * nx)
        if pack_binary:
            data = np.unpackbits(np.fromstring(payload, data_type))[:len(zrange) * ny * nx]
        else:
            data = np.fromstring(payload, data_type)
        # convert the payload into actual 3D data
        data = np.reshape(data.astype(data_type), (len(zrange), ny, nx), order='C')
        f.close()
    # HP 10/2013 start using proper [x,y,z] data ordering
    data_xyz = data.transpose(2, 1, 0)
    return data_xyz


def rawmar_read(image_name, size, verbose=False):
    '''Read a square 2D image plate MAR image.

    These binary images are typically obtained from the marcvt utility.

    .. note::

       This method assume Big endian byte order.
    '''
    data = HST_read(image_name, dims=(1, size, size), header=4600,
                    data_type=np.uint16, verbose=verbose)[:, :, 0]
    return data


def HST_write(data, file_name, mode='w', verbose=True, pack_binary=False):
    '''Write data as a raw binary file.

    This function write a (x,y,z) 3D dataset to the disk. The actual data type is used, you can convert your data array 
    on the fly using data.astype if you want to change the type. 
    The file is written as a Z-stack. It means that the first nx*ny bytes written represent the first slice and so on...
    For binary data files (stored in memory as integer or bool data type), binary packing mode can be activated which 
    stores 8 values on each byte (saving 7/8 of the disk space).

    A .info file containing the volume size and data type is also written.

    :param data: the 3d array to write to the disk in [x, y, z] form.
    :param str file_name: the name of the file to write, including file extension.
    :param char mode: file write mode, change to 'a' to append to a file.
    :param bool verbose: flag to activate verbose mode.
    :param bool pack_binary: flag to activate binary packing.
    '''
    if data.dtype == np.bool:
        print('casting bool array to uint8, you may consider using binary packing to save disk space.')
        data = data.astype(np.uint8)
    (nx, ny, nz) = data.shape
    if verbose:
        print('opening %s for writing in mode %s' % (file_name, mode))
        print('volume size is %dx%dx%d' % (nx, ny, nz))
        print('data type is %s' % data.dtype)
    f = open(file_name, mode + 'b')
    # HP 11/2013 swap axes according to read function
    if pack_binary:
        s = np.packbits(data.astype(np.uint8).transpose(2, 1, 0)).tostring()
    else:
        s = np.ravel(data.transpose(2, 1, 0)).tostring()
    f.write(s)
    f.close()
    if verbose:
        print('writing .info file')
    f = open(file_name + '.info', mode)
    f.write('! PyHST_SLAVE VOLUME INFO FILE\n')
    f.write('NUM_X = %4d\n' % nx)
    f.write('NUM_Y = %4d\n' % ny)
    f.write('NUM_Z = %4d\n' % nz)
    if pack_binary:
        f.write('DATA_TYPE = PACKED_BINARY\n')
    else:
        f.write('DATA_TYPE = %s\n' % data.dtype)
    f.close()
    if verbose:
        print('done with writing')


def recad_vol(vol_filename, min, max, verbose=False):
    '''Recad a 32 bit vol file into 8 bit raw file.

    This function reads a 3D volume file into a numpy float32 array and
    applies the `recad` function with the [min, max] range. The result is
    saved into a .raw file with the same name as the input file.

    In verbose mode, a piture to compare mid slices in the two volumes
    and another one to compare the histograms are saved.

    .. note::

       To read the vol file, the presence of a .info file is
       assumed, see `HST_read`.

    *Parameters*

    **vol_filename**: the path to the binary vol file.

    **min**: value to use as the minimum (will be 0 in the casted array).

    **max**: value to use as the maximum (will be 255 in the casted array).

    **verbose**: activate verbose mode (False by default).
    '''
    prefix = vol_filename[:-4]
    infos = HST_info(vol_filename + '.info')
    vol_size = [infos['x_dim'], infos['y_dim'], infos['z_dim']]
    data = HST_read(vol_filename, type=np.float32)
    data_uint8 = recad(data, min, max)
    if verbose:
        plt.figure(1, figsize=(10, 5))
        plt.subplot(121)
        plt.imshow(data[:, :, vol_size[2] // 2])
        plt.title('float image')
        plt.axis('off')
        plt.subplot(122)
        plt.imshow(data_uint8[:, :, vol_size[2] // 2])
        plt.title('uint8 image')
        plt.axis('off')
        plt.savefig('%sslices.pdf' % (prefix), format='pdf')
        plt.figure(2)
        plt.clf()
        plt.subplot(211)
        plt.title('Gray level histograms from float to uint8')
        n, bins, patches = plt.hist(data.ravel(), bins=256, histtype='stepfilled', facecolor='green')
        plt.figure(2)
        plt.subplot(212)
        plt.hist(data_uint8.ravel(), bins=256, histtype='stepfilled', facecolor='green')
        plt.savefig('%shist.pdf' % (prefix), format='pdf')
    HST_write(data_uint8, '%s.raw' % prefix)


def Vtk_write(data, fname):
    '''Write a data array into old style (V3.0) VTK format.

    An ascii header is written to which the binary data is appended.

    .. note::

       The header assumes uint8 data type.
    '''
    (nz, ny, nx) = data.shape
    print('opening', fname, 'for writing')
    print('volume size is %d x %d x %d' % (nx, ny, nz))
    # write header
    f = open(fname, 'w')
    f.write('# vtk DataFile Version3.0\n')
    f.write(fname[:-4] + '\n')
    f.write('BINARY\n');
    f.write('DATASET STRUCTURED_POINTS\n');
    f.write('DIMENSIONS ' + str(nx) + ' ' + str(ny) + ' ' + str(nz) + '\n');
    f.write('SPACING 1.0 1.0 1.0\n');
    f.write('ORIGIN 0.0 0.0 0.0\n');
    f.write('POINT_DATA ' + str(nx * ny * nz) + '\n');
    f.write('SCALARS volume_scalars unsigned_char 1\n');
    f.write('LOOKUP_TABLE default\n')
    f.close()
    # append binary data
    f = open(fname, 'ab')
    s = np.ravel(data).tostring()
    f.write(s)
    f.close()
    print('done with writing')



"""The dct module provide helpers functions to work with experimental diffraction contrast tomography data.
"""
import os
import h5py
import numpy as np
from scipy import ndimage
from matplotlib import pyplot as plt, cm
from pymicro.xray.experiment import ForwardSimulation
from pymicro.crystal.lattice import HklPlane
from pymicro.xray.xray_utils import lambda_keV_to_nm, radiograph, radiographs
from pymicro.crystal.microstructure import Grain, Orientation


class DctForwardSimulation(ForwardSimulation):
    """Class to represent a Forward Simulation."""

    def __init__(self, verbose=False):
        super(DctForwardSimulation, self).__init__('dct', verbose=verbose)
        self.hkl_planes = []
        self.check = 1  # grain id to display infos in verbose mode
        self.omegas = None
        self.reflections = []

    def set_hkl_planes(self, hkl_planes):
        self.hkl_planes = hkl_planes

    def set_diffracting_famillies(self, hkl_list):
        """Set the list of diffracting hk planes using a set of families."""
        symmetry = self.exp.get_sample().get_material().get_symmetry()
        hkl_planes = []
        for hkl in hkl_list:
            # here we set include_friedel_pairs to False as we take it into account in the calculation
            planes = HklPlane.get_family(hkl, include_friedel_pairs=True, crystal_structure=symmetry)
            for plane in planes:  # fix the lattice
                plane.set_lattice(self.exp.get_sample().get_material())
            hkl_planes.extend(planes)
        self.set_hkl_planes(hkl_planes)


    def setup(self, omega_step):
        """Setup the forward simulation."""
        assert self.exp.source.min_energy == self.exp.source.max_energy  # monochromatic case
        lambda_keV = self.exp.source.max_energy
        self.omegas = np.linspace(0.0, 360.0, num=int(360.0 / omega_step), endpoint=False)
        self.reflections = []
        for omega in self.omegas:
            self.reflections.append([])
        for g in self.exp.sample.microstructure.grains:
            for plane in self.hkl_planes:
                (h, k, i, l) = HklPlane.three_to_four_indices(*plane.miller_indices())
                try:
                    (w1, w2) = g.dct_omega_angles(plane, lambda_keV, verbose=False)
                except ValueError:
                    if self.verbose:
                        print('plane {} does not fulfil the Bragg condition for grain {:d}'.format((h, k, i, l), g.id))
                    continue
                # add angles for Friedel pairs
                w3 = (w1 + 180.) % 360
                w4 = (w2 + 180.) % 360
                if self.verbose and g.id == self.check:
                    print('grain %d, angles for plane %d%d%d: w1=%.3f and w2=%.3f | delta=%.1f' % (g.id, h, k, l, w1, w2, w1-w2))
                    print('(%3d, %3d, %3d, %3d) -- %6.2f & %6.2f' % (h, k, i, l, w1, w2))
                self.reflections[int(w1 / omega_step)].append([g.id, (h, k, l)])
                self.reflections[int(w2 / omega_step)].append([g.id, (h, k, l)])
                self.reflections[int(w3 / omega_step)].append([g.id, (-h, -k, -l)])
                self.reflections[int(w4 / omega_step)].append([g.id, (-h, -k, -l)])

    def load_grain(self, gid=1):
        print('loading grain from file 4_grains/phase_01/grain_%04d.mat' % gid)
        with h5py.File(os.path.join(self.exp.get_sample().data_dir, '4_grains/phase_01/grain_%04d.mat' % gid)) as gmat:
            g = Grain(gid, Orientation.from_rodrigues(gmat['R_vector'][()]))
            g.om_exp = gmat['om_exp'][0, :]
            g.uv_exp = gmat['uv_exp'][:, :]
            g.center = gmat['center'][:, 0]
            try:
                ref_included = gmat['proj/included'][0][0]
                g.included = gmat[ref_included][0, :]
                ref_ondet = gmat['proj/ondet'][0][0]
                g.ondet = gmat[ref_ondet][0, :]
                # grab the projection stack
                ref_stack = gmat['proj']['stack'][0][0]
                g.stack_exp = gmat[ref_stack][()].transpose(1, 2, 0)  # now in [ndx, u, v] form
                g.hklsp = gmat['allblobs/hklsp'][:, :]
            except AttributeError:
                # classic file organization
                g.included = gmat['proj/included'][0, :]
                g.ondet = gmat['proj/ondet'][0, :]
                g.stack_exp = gmat['proj/stack'][()].transpose(1, 2, 0)  # now in [ndx, u, v] form
                # for this data set, we have to hack around the DCT + TT work in progress
                ref_hklsp = gmat['allblobs/hklsp'][()][0][0]
                g.hklsp = gmat[ref_hklsp][:, :]
        self.grain = g
        if self.verbose:
            print('experimental proj stack shape: {}'.format(g.stack_exp.shape))

    def grain_projection_image(self, g_uv, g_proj):
        """Produce a 2D image placing all diffraction spots of a given grain at their respective position on the detector.

        Spots outside the detector are are skipped while those only partially on the detector are cropped accordingly.

        :param g_proj: image stack of the diffraction spots. The first axis is so that g_proj[0] is the first spot \
        the second axis is the horizontal coordinate of the detector (u) and the third axis the vertical coordinate \
        of the detector (v).
        :param g_uv: list or array of the diffraction spot position.
        :returns: a 2D composite image of all the diffraction spots.
        """
        print(len(g_proj), g_uv.shape[1])
        assert len(g_proj) == g_uv.shape[1]
        image = np.zeros(self.exp.get_active_detector().get_size_px())
        for i in range(len(g_proj)):
            spot = g_proj[i]
            if self.verbose:
                print('i={0}, size of spot: {1}'.format(i, spot.shape))
                print('placing diffraction spot at location {0}'.format(g_uv[:, i]))
            add_to_image(image, spot, g_uv[:, i], self.verbose)
        return image

    def grain_projection_exp(self, gid=1):
        """Produce a composite image with all the experimental diffraction spots of this grain on the detector.

        :param int gid: the number of the selected grain.
        :returns: a 2D composite image of all the diffraction spots.
        """
        #self.grain = self.exp.get_sample().get_microstructure().get_grain(gid)
        if not hasattr(self, 'grain') or self.grain.id != gid:
            # load the corresponding grain
            self.load_grain(gid=gid)
        return self.grain_projection_image(self.grain.uv_exp, self.grain.stack_exp)

    def grain_projection_simulation(self, gid=1, data=None):
        """Function to compute all the grain projection in DCT geometry and create a composite image."""
        print('forward simulation of grain %d' % gid)
        detector = self.exp.get_active_detector()
        lambda_keV = self.exp.source.max_energy
        lambda_nm = lambda_keV_to_nm(lambda_keV)
        X = np.array([1., 0., 0.]) / lambda_nm
        lattice = self.exp.get_sample().get_material()

        if not hasattr(self, 'grain'):
            # load the corresponding grain
            self.load_grain(gid=gid)
        from scipy import ndimage
        if data is None:
            grain_ids = self.exp.get_sample().get_grain_ids()
            print('binarizing grain %d' % gid)
            data = np.where(grain_ids[ndimage.find_objects(grain_ids == gid)[0]] == gid, 1, 0)
        print('shape of binary grain is {}'.format(data.shape))

        # compute all the omega values
        print('simulating diffraction spot positions on the detector')
        omegas = np.zeros(2 * len(self.hkl_planes))
        g_uv = np.zeros((2, 2 * len(self.hkl_planes)))
        for i, plane in enumerate(self.hkl_planes):
            #print(plane.miller_indices())
            try:
                w1, w2 = self.grain.dct_omega_angles(plane, lambda_keV, verbose=False)
            except ValueError:
                # plane does not fulfil the Bragg condition
                continue
            omegas[2 * i] = w1
            omegas[2 * i + 1] = w2
            for j in range(2):
                omega = omegas[2 * i + j]
                omegar = omega * np.pi / 180
                R = np.array([[np.cos(omegar), -np.sin(omegar), 0], [np.sin(omegar), np.cos(omegar), 0], [0, 0, 1]])
                gt = self.grain.orientation_matrix().transpose()
                G = np.dot(R, np.dot(gt, plane.scattering_vector()))
                K = X + G
                # position of the grain at this rotation angle
                g_pos_rot = np.dot(R, self.grain.center)
                pg = detector.project_along_direction(K, g_pos_rot)
                (up, vp) = detector.lab_to_pixel(pg)[0]
                g_uv[:, 2 * i + j] = up, vp
        # build the 3d projection stack at once
        print('building grain projections stack')
        stack_sim = radiographs(data, omegas)
        stack_sim = stack_sim.transpose(2, 0, 1)[:, :, ::-1]  # (u, v) axes correspond to (Y, -Z) for DCT detector
        return self.grain_projection_image(g_uv, stack_sim)


    def dct_projection(self, omega, include_direct_beam=True, att=5):
        """Function to compute a full DCT projection at a given omega angle."""
        if len(self.reflections) == 0:
            print('empty list of reflections, you should run the setup function first')
            return None
        grain_ids = self.exp.get_sample().get_grain_ids()
        detector = self.exp.get_active_detector()
        lambda_keV = self.exp.source.max_energy
        lattice = self.exp.get_sample().get_material()
        index = np.argmax(self.omegas > omega)
        dif_grains = self.reflections[index - 1]  # grains diffracting between omegas[index - 1] and omegas[index]
        # intialize image result
        full_proj = np.zeros(detector.get_size_px(), dtype=np.float)
        lambda_nm = lambda_keV_to_nm(lambda_keV)
        omegar = omega * np.pi / 180
        R = np.array([[np.cos(omegar), -np.sin(omegar), 0], [np.sin(omegar), np.cos(omegar), 0], [0, 0, 1]])

        if include_direct_beam:
            # add the direct beam part by computing the radiograph of the sample without the diffracting grains
            data_abs = np.where(grain_ids > 0, 1, 0)
            for (gid, (h, k, l)) in dif_grains:
                mask_dif = (grain_ids == gid)
                data_abs[mask_dif] = 0  # remove this grain from the absorption
            proj = radiograph(data_abs, omega)[:, ::-1]  # (u, v) axes correspond to (Y, -Z) for DCT detector
            add_to_image(full_proj, proj / att, np.array(full_proj.shape) // 2)

        # add diffraction spots
        X = np.array([1., 0., 0.]) / lambda_nm
        for (gid, (h, k, l)) in dif_grains:
            grain_data = np.where(grain_ids == gid, 1, 0)
            if np.sum(grain_data) < 1:
                print('skipping grain %d' % gid)
                continue
            local_com = np.array(ndimage.measurements.center_of_mass(grain_data, grain_ids))
            print('local center of mass (voxel): {0}'.format(local_com))
            g_center_mm = detector.get_pixel_size() * (local_com - 0.5 * np.array(grain_ids.shape))
            print('center of mass (voxel): {0}'.format(local_com - 0.5 * np.array(grain_ids.shape)))
            print('center of mass (mm): {0}'.format(g_center_mm))
            # compute scattering vector
            gt = self.exp.get_sample().get_microstructure().get_grain(gid).orientation_matrix().transpose()
            p = HklPlane(h, k, l, lattice)
            G = np.dot(R, np.dot(gt, p.scattering_vector()))
            K = X + G
            # position of the grain at this rotation angle
            g_pos_rot = np.dot(R, g_center_mm)
            pg = detector.project_along_direction(K, g_pos_rot)
            (up, vp) = detector.lab_to_pixel(pg)[0]
            if self.verbose:
                print('\n* gid=%d, (%d,%d,%d) plane, angle=%.1f' % (gid, h, k, l, omega))
                print('diffraction vector:', K)
                print('postion of the grain at omega=%.1f is ' % omega, g_pos_rot)
                print('up=%d, vp=%d for plane (%d,%d,%d)' % (up, vp, h, k, l))
            data_dif = grain_data[ndimage.find_objects(grain_ids == gid)[0]]
            proj_dif = radiograph(data_dif, omega)  # (Y, Z) coordinate system
            add_to_image(full_proj, proj_dif[:, ::-1], (up, vp), self.verbose)  # (u, v) axes correspond to (Y, -Z)
        return full_proj


def add_to_image(image, inset, uv, verbose=False):
    """Add an image to another image at a specified position.

    The inset image may be of any size and may only overlap partly on the overall image depending on the location
    specified. In such a case, the inset image is cropped accordingly.

    :param np.array image: the master image taht will be modified.
    :param np.array inset: the inset to add to the image.
    :param tuple uv: the location (center) where to add the inset in the form (u, v).
    :param bool verbose: activate verbose mode (False by default).
    """
    # round the center to the closest integer value
    u = int(uv[0])
    v = int(uv[1])
    spot_size = inset.shape
    u_start = 0
    u_end = spot_size[0]
    v_start = 0
    v_end = spot_size[1]
    # check bounds for spots that may be completely or partly outside the image
    if (u + spot_size[0] // 2 < 0) or (u - spot_size[0] // 2 > image.shape[0] - 1) or (
                    v + spot_size[1] // 2 < 0) or (v - spot_size[1] // 2 > image.shape[1] - 1):
        if verbose:
            print('skipping this spot which is outside the detector area')
        return None  # spot is completely outside the detector area
    if u - spot_size[0] // 2 < 0:
        u_start = int(spot_size[0] // 2 - u + 1)
    elif u - spot_size[0] // 2 + spot_size[0] > image.shape[0] - 1:
        u_end = int(image.shape[0] - (u - spot_size[0] // 2))
    if v - spot_size[1] // 2 < 0:
        v_start = int(spot_size[1] // 2 - v + 1)
    elif v - spot_size[1] // 2 + spot_size[1] > image.shape[1] - 1:
        v_end = int(image.shape[1] - (v - spot_size[1] // 2))
    # now add spot to the image
    image[u - spot_size[0] // 2 + u_start: u - spot_size[0] // 2 + u_end,
    v - spot_size[1] // 2 + v_start: v - spot_size[1] // 2 + v_end] \
        += inset[u_start:u_end, v_start:v_end]


def merge_dct_scans(scan_list, samtz_list, use_mask=False, overlap=-1, root_dir='.', write_to_h5=True):
    """Merge two DCT scans.

    This function build a `Microstructure` instance for each DCT scan and calls merge_microstructures.
    The overlap can be deduced from the samtz values or specified directly.

    :param list scan_list: a list with the two DCT scan names.
    :param list samtz_list: a list with the two samtz value (the order should match the scan names).
    :param bool use_mask: a flag to also merge the absorption masks.
    :param int overlap: the value to use for the overlap if not computed automatically.
    :param str root_dir: the root data folder.
    :param bool write_to_h5: flag to write the result of the merging operation to an HDF5 file.
    :return: A new `Microstructure` instance of the 2 merged scans.
    """
    from pymicro.crystal.microstructure import Microstructure
    import numpy as np
    import os
    import h5py

    scan_shapes = []  # warning, shapes will be in (z, y, x) form
    micros = []

    for scan in scan_list:
        scan_path = os.path.join(root_dir, scan, '5_reconstruction', 'phase_01_vol.mat')
        with h5py.File(scan_path) as f:
            scan_shapes.append(f['vol'].shape)
            print(f['vol'].shape)

    # figure out the maximum cross section
    max_shape = np.array(scan_shapes).max(axis=0)[[2, 1, 0]]

    for scan in scan_list:
        # read microstructure for this scan
        dct_analysis_dir = os.path.join(root_dir, scan)
        print('processing scan %s' % scan)
        micro = Microstructure.from_dct(data_dir=dct_analysis_dir)
        print('voxel_size is {}'.format(micro.voxel_size))

        # pad both grain map and mask
        print('max shape is {}'.format(max_shape))
        print('vol shape is {}'.format(micro.grain_map.shape))
        offset = max_shape - micro.grain_map.shape
        offset[2] = 0  # do not pad along Z
        padding = [(o // 2, max_shape[0] - micro.grain_map.shape[0] - o // 2) for o in offset]
        print('padding is {}'.format(padding))
        micro.grain_map = np.pad(micro.grain_map, padding, mode='constant')
        print('has mask ? {}'.format(hasattr(micro, 'mask')))
        if use_mask:
            micro.mask = np.pad(micro.mask, padding, mode='constant')
        elif hasattr(micro, 'mask'):
            print('deleting mask attribute since we do not want to use it')
            delattr(micro, 'mask')
        micros.append(micro)

    # find out the overlap region (based on the difference in samtz)
    overlap_from_samtz = int((samtz_list[1] + scan_shapes[1][0] // 2 * micros[1].voxel_size) / micros[1].voxel_size
                   - (samtz_list[0] - scan_shapes[0][0] // 2 * micros[0].voxel_size) / micros[0].voxel_size)
    print('vertical overlap deduced from samtz positions is %d voxels' % overlap_from_samtz)
    if overlap < 0:
        overlap = overlap_from_samtz
    print('using an actual overlap of %d voxels' % overlap)

    # we have prepared the 2 microstructures, now merge them
    merged_micro = Microstructure.merge_microstructures(micros, overlap, plot=True)

    if write_to_h5:
        # write the result
        merged_micro.to_h5()

    return merged_micro


def all_dif_spots(g_proj, g_uv, verbose=False):
    """Produce a 2D image placing all diffraction spots at their respective position on the detector.

    Spots outside the detector are are skipped while those only partially on the detector are cropped accordingly.

    :param g_proj: image stack of the diffraction spots. The first axis is so that g_proj[0] is the first spot \
    the second axis is the horizontal coordinate of the detector (u) and the third axis the vertical coordinate \
    of the detector (v).
    :param g_uv: list or array of the diffraction spot position.
    :param bool verbose: activate verbose mode (False by default).
    :returns: a 2D composite image of all the diffraction spots.
    """
    # TODO add a detector object to account for the image size and the position of the direct beam
    image = np.zeros((2048, 2048), dtype=g_proj.dtype)
    print(g_proj.shape[0], len(g_uv))
    assert g_proj.shape[0] == len(g_uv)
    for i in range(g_proj.shape[0]):
        spot = g_proj[i]
        if verbose:
            print('i={0}, size of spot: {1}'.format(i, spot.shape))
            print('placing diffraction spot at location {0}'.format(g_uv[i]))
        add_to_image(image, spot, g_uv[i], verbose)
    return image


def plot_all_dif_spots(gid, detector, hkl_miller=None, uv=None, lattice=None, lambda_keV=None,
                       spots=True, dif_spots_image=None, max_value=None, positions=True, debye_rings=True, suffix=''):
    plt.figure(figsize=(13, 8))
    # plt.figure()
    if spots and dif_spots_image is not None:
        if not max_value:
            max_value = dif_spots_image.max()
        plt.imshow(dif_spots_image.T, cmap=cm.gray, vmin=0, vmax=max_value)
    families = []
    indices = []
    colors = 'crbgmy'  # crbgmycrbgmycrbgmycrbgmy'  # use a cycler here
    t = np.linspace(0.0, 2 * np.pi, num=37)
    if positions or debye_rings:
        if hkl_miller is None:
            raise ValueError(
                'The list of miller indices of each reflection must be provided using variable g_hkl_miller')
        for i, (h, k, l) in enumerate(hkl_miller):
            l = [abs(h), abs(k), abs(l)]
            # l.sort()  # miller indices are now sorted, should use the lattice symmetry here
            family_name = '%d%d%d' % (l[0], l[1], l[2])
            if families.count(family_name) == 0:
                families.append(family_name)
                indices.append(i)
        indices.append(len(hkl_miller))
        print(families, indices)
        # now plot each family
        for i in range(len(families)):
            family = families[i]
            c = colors[i % len(colors)]
            if positions and uv is not None:
                plt.plot(uv[indices[i]:indices[i + 1], 0], uv[indices[i]:indices[i + 1], 1],
                         's', color=c, label=family)
            if debye_rings and lattice is not None and lambda_keV:
                theta = HklPlane(int(family[0]), int(family[1]), int(family[2]), lattice).bragg_angle(lambda_keV)
                L = detector.ref_pos[0] / detector.pixel_size * np.tan(2 * theta)  # 2 theta distance on the detector
                print('2theta = %g, L = %g' % (2 * theta * 180 / np.pi, L))
                plt.plot(0.5 * detector.size[0] + L * np.cos(t), 0.5 * detector.size[1] + L * np.sin(t), '--', color=c)
    plt.title('grain %d diffraction spot locations on the detector' % gid)
    # plt.legend(numpoints=1, loc='center')
    plt.legend(numpoints=1, ncol=2, bbox_to_anchor=(1.02, 1), loc=2)
    # plt.axis('equal')
    plt.axis([0, 2047, 2047, 0])
    plt.savefig('g%d%s_difspot_positions.pdf' % (gid, suffix))


def output_tikzpicture(proj_dif, omegas, gid=1, d_uv=[0, 0], suffix=''):
    axes = ['horizontal', 'vertical']
    for i, d in enumerate(d_uv):
        if d:
            # pad the corresponding axis if needed
            p = (d - proj_dif.shape[i + 1]) // 2
            if p > 0:
                print('padding %s axis with %d zeros' % (axes[i], p))
                if i == 0:
                    proj_dif_pad = np.pad(proj_dif, ((0, 0), (p, p), (0, 0)), 'constant')
                else:
                    proj_dif_pad = np.pad(proj_dif, ((0, 0), (0, 0), (p, p)), 'constant')
        else:
            d_uv[i] = proj_dif.shape[i + 1]
    print('output proj images will be {0}'.format(d_uv))
    N = proj_dif.shape[0]
    for i in range(N):
        proj_path = os.path.join('proj_dif', 'g%d_proj%s_%02d.png' % (gid, suffix, i))
        # proj_dif is in (u, v) form so as usual, we need to save the transposed array with imsave
        plt.imsave(proj_path,
                   proj_dif[i, (proj_dif.shape[1] - d_uv[0]) // 2:(proj_dif.shape[1] + d_uv[0]) // 2,
                   (proj_dif.shape[2] - d_uv[1]) // 2:(proj_dif.shape[2] + d_uv[1]) // 2].T,
                   cmap=cm.gray, origin='upper')
    # HST_write(proj_dif, 'g4_proj_stack.raw')

    # image ratio is 4/3 so we will use a n x m matrix with
    L = 10.0  # cm
    H = 3 * L / 4  # cm
    n = 0
    while (3. / 4 * n ** 2) < N:
        n += 1
    m = int(3. * n / 4)
    # size of a picture (assuming square)
    l = L / n
    tex_path = os.path.join('proj_dif', 'g%d_proj%s.tex' % (gid, suffix))
    print('building a tikzpicture with {0}x{1} miniatures, output file {2}'.format(n, m, tex_path))
    f = open(tex_path, 'w')
    f.write('\\documentclass{article}')
    f.write('\\usepackage[latin1]{inputenc}\n')
    f.write('\\usepackage{tikz}\n')
    f.write('\\usetikzlibrary{shapes,arrows}\n')
    f.write('\\usepackage{xcolor}\n')
    f.write('\\pagecolor{black}\n')
    f.write('\\usepackage[active,tightpage]{preview}\n')
    f.write('\\PreviewEnvironment{tikzpicture}\n')
    f.write('\\setlength\PreviewBorder{0pt}\n')
    f.write('\\begin{document}\n')
    f.write('\\pagestyle{empty}\n')
    f.write('\\sffamily\n')
    f.write('\\fontsize{5}{6}\n')
    f.write('\\begin{tikzpicture}[white]\n')
    # f.write(
    #    '\\filldraw[black] (%.3f,%.3f) rectangle (%.3f,%.3f);\n' % (-0.5 * l, 0.5 * l, L - 0.5 * l, -m * l + 0.5 * l))
    for j in range(m):
        Y = -j * l
        for i in range(n):
            X = i * l
            index = j * n + i
            if index >= N:
                continue  # skip last incomplete line
            f.write('\\node at (%.2f,%.2f) {\includegraphics[width=%.2fcm]{g%d_proj%s_%02d.png}};\n' % (
                X, Y, l, gid, suffix, index))
            # f.write('\\node at (%.2f,%.2f) {$\omega=%.1f^\circ$};\n' % (X, Y - 0.5 * l, omegas[index]))
            f.write('\\node at (%.2f,%.2f) {$%.1f^\circ$};\n' % (X, Y - 0.5 * l, omegas[index]))
    f.write('\\end{tikzpicture}\n')
    f.write('\\end{document}\n')
    f.close()


def tt_stack(scan_name, data_dir='.', save_edf=False, TOPO_N=-1, dark_factor=1.):
    """Build a topotomography stack from raw detector images.

    The number of image to sum for a topograph can be determined automatically from the total number of images present
    in `data_dir` or directly specified using the variable `TOPO_N`.

    :param str scan_name: the name of the scan to process.
    :param str data_dir: the path to the data folder.
    :param bool save_edf: flag to save the tt stack as an EDF file.
    :param int TOPO_N: the number of images to sum for a topograph.
    :param float dark_factor: a multiplicative factor for the dark image.
    """
    from pymicro.file.file_utils import edf_read, edf_write
    if TOPO_N < 0:
        import glob
        # figure out the number of frames per topograph TOPO_N
        n_frames = len(glob.glob(os.path.join(data_dir, scan_name, '%s*.edf' % scan_name)))
        TOPO_N = int(n_frames / 90)
    print('number of frames to sum for a topograph = %d' % TOPO_N)

    # parse the info file
    f = open(os.path.join(data_dir, scan_name, '%s.info' % scan_name))
    infos = dict()
    for line in f.readlines():
        tokens = line.split('=')
        # convert the value into int/float/str depending on the case
        try:
            value = int(tokens[1].strip())
        except ValueError:
            try:
                value = float(tokens[1].strip())
            except ValueError:
                value = tokens[1].strip()
        infos[tokens[0]] = value
    print(infos)

    # load dark image
    dark = dark_factor * edf_read(os.path.join(data_dir, scan_name, 'darkend0000.edf'))

    # build the stack by combining individual images
    tt_stack = np.empty((infos['TOMO_N'], infos['Dim_1'], infos['Dim_2']))
    print(tt_stack[0].shape)
    for n in range(int(infos['TOMO_N'])):
        print('building topograph %d' % (n + 1))
        topograph = np.zeros((infos['Dim_1'], infos['Dim_2']))
        offset = TOPO_N * n
        for i in range(TOPO_N):
            index = offset + i + 1
            frame_path = os.path.join(data_dir, scan_name, '%s%04d.edf' % (scan_name, index))
            im = edf_read(frame_path) - dark
            topograph += im
        tt_stack[n] = topograph
    tt_stack = tt_stack.transpose((1, 2, 0))
    print('done')

    # save the data as edf if needed
    if save_edf:
        edf_write(tt_stack, os.path.join(data_dir, '%sstack.edf' % scan_name))
    return tt_stack


"""The texture module provide some utilities to generate, analyse and plot crystallographic textures.
"""
import numpy as np
from pymicro.crystal.lattice import Symmetry, Lattice, HklPlane, SlipSystem
from pymicro.crystal.microstructure import Orientation, Grain, Microstructure
from matplotlib import pyplot as plt, colors, cm


class PoleFigure:
    """A class to handle pole figures.

    A pole figure is a popular tool to plot multiple crystal orientations,
    either in the sample coordinate system (direct pole figure) or
    alternatively plotting a particular direction in the crystal
    coordinate system (inverse pole figure).
    """

    def __init__(self, microstructure=None, lattice=None, axis='Z', hkl='111',
                 proj='stereo', verbose=False):
        """
        Create an empty PoleFigure object associated with an empty Microstructure.

        :param microstructure: the :py:class:`~pymicro.crystal.microstructure.Microstructure` containing the collection of orientations to plot (None by default).
        :param lattice: the crystal :py:class:`~pymicro.crystal.lattice.Lattice`.
        :param str axis: the pole figure axis ('Z' by default), vertical axis in the direct pole figure and direction plotted on the inverse pole figure.

        .. warning::

           Any crystal structure is now supported (you have to set the proper
           crystal lattice) but it has only really be tested for cubic.

        :param str hkl: slip plane family ('111' by default)
        :param str proj: projection type, can be either 'stereo' (default) or 'flat'
        :param bool verbose: verbose mode (False by default)
        """
        self.proj = proj
        self.axis = axis
        self.map_field = None
        if microstructure:
            self.microstructure = microstructure
        else:
            self.microstructure = Microstructure()
        if lattice:
            self.lattice = lattice
        else:
            self.lattice = Lattice.cubic(1.0)
        self.family = None
        self.poles = []
        self.set_hkl_poles(hkl)
        self.verbose = verbose
        self.resize_markers = False
        self.mksize = 50
        self.pflegend = False
        self.x = np.array([1., 0., 0.])
        self.y = np.array([0., 1., 0.])
        self.z = np.array([0., 0., 1.])

        # list all crystal directions
        self.c001s = np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]], dtype=np.float)
        self.c011s = np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0], [0, -1, 1], [-1, 0, 1], [-1, 1, 0]],
                              dtype=np.float) / np.sqrt(2)
        self.c111s = np.array([[1, 1, 1], [-1, -1, 1], [1, -1, 1], [-1, 1, 1]], dtype=np.float) / np.sqrt(3)

    def get_orientations(self):
        """Get the list of orientations in the PoleFigure.
        
        :return: a list of `Orientation` instances.
        """
        return [grain.orientation for grain in self.microstructure.grains]

    def set_hkl_poles(self, hkl='111'):
        """Set the pole (aka hkl planes) list to to use in the `PoleFigure`.

        The list of poles can be given by the family type or directly by a list of `HklPlanes` objects.

        :params str/list hkl: slip plane family ('111' by default)
        """
        if type(hkl) is str:
            self.family = hkl  # keep a record of this
            hkl_planes = self.lattice.get_hkl_family(self.family)
        elif type(hkl) is list:
            self.family = None
            hkl_planes = hkl
        self.poles = hkl_planes  #[p.normal() for p in hkl_planes]

    def set_map_field(self, field_name, field=None, field_min_level=None, field_max_level=None, lut='hot'):
        '''Set the PoleFigure to color poles with the given field.

        This method activates a mode where each symbol in the pole figure
        is color coded with respect to a field, which can be either the
        grain id, or a given field given in form of a list. If the grain
        volume or strain. For the grain id, the color is set according the
        each grain id in the :py:class:`~pymicro.crystal.microstructure.Microstructure`
        and the :py:meth:`~pymicro.crystal.microstructure.rand_cmap` function.
        For a given field, the color is set from the lookup table and
        according to the value in the given list. The list must contain a
        record for each grain. Minimum and maximum value to map the field
        values and the colors can be specify, if not they are directly taken
        as the min() and max() of the field.

        :param str field_name: The field name, could be 'grain_id', or any other name describing the field.
        :param list field: A list containing a record for each grain.
        :param float field_min_level: The minimum value to use for this field.
        :param float field_max_level: The maximum value to use for this field.
        :param str lut: A string describing the colormap to use (among matplotlib ones available).
        :raise ValueError: If the given field does not contain enough values.
        '''
        self.map_field = field_name
        self.lut = lut
        if field_name == 'grain_id':
            self.field = [g.id for g in self.microstructure.grains]
        elif field_name == 'ipf':
            self.field = [g.id for g in self.microstructure.grains]
        else:
            if len(field) < len(self.microstructure.grains):
                raise ValueError('The field must contain a record for each grain in the microstructure')
            self.field = field
            if not field_min_level:
                self.field_min_level = field.min()
            else:
                self.field_min_level = field_min_level
            if not field_max_level:
                self.field_max_level = field.max()
            else:
                self.field_max_level = field_max_level

    def plot_pole_figures(self, plot_sst=True, display=True, save_as='pdf'):
        '''Plot and save a picture with both direct and inverse pole figures.

        :param bool plot_sst: controls wether to plot the full inverse pole \
        figure or only the standard stereographic triangle (True by default).
        :param bool display: display the plot if True, else save a picture \
        of the pole figures (True by default)
        :param str save_as: File format used to save the image such as pdf \
        or png ('pdf' by default)

        ::

          micro = Microstructure(name = 'AlLi_sam8')
          micro.grains.append(Grain(11, Orientation.from_euler(np.array([262.364, 16.836, 104.691]))))
          Al_fcc = Lattice.face_centered_cubic(0.405) # not really necessary since default lattice is cubic
          pf = PoleFigure(microstructure=micro, proj='stereo', lattice=Al_fcc, hkl='111')
          pf.mksize = 12
          pf.set_map_field('grain_id')
          pf.pflegend = True # this works well for a few grains
          pf.plot_pole_figures()

        .. figure:: _static/AlLi_sam8_pole_figure.png
            :width: 750 px
            :height: 375 px
            :alt: AlLi_sam8_pole_figure
            :align: center

            A 111 pole figure plotted for a single crystal orientation.
        '''
        fig = plt.figure(figsize=(10, 5))
        # direct PF
        ax1 = fig.add_subplot(121, aspect='equal')
        self.plot_pf(ax=ax1, mk='o', ann=False)
        # inverse PF
        ax2 = fig.add_subplot(122, aspect='equal')
        if plot_sst:
            self.plot_sst(ax=ax2)
        else:
            self.plot_ipf(ax=ax2)
        if display:
            plt.show()
        else:
            plt.savefig('%s_pole_figure.%s' % (self.microstructure.name, save_as), format=save_as)

    def plot_crystal_dir(self, c_dir, **kwargs):
        '''Function to plot a crystal direction on a pole figure.

        :param c_dir: A vector describing the crystal direction.
        :param dict kwargs: a dictionnary of keyword/values to control the plot, it should at least contain a reference
        to a pyplot axes to draw the pole using keywors 'ax'.
        :raise ValueError: if the projection type is not supported
        '''
        if c_dir[2] < 0:
            c_dir *= -1  # make unit vector have z>0
        if self.proj == 'flat':
            cp = c_dir
        elif self.proj == 'stereo':
            c = c_dir + self.z
            c /= c[2]  # SP'/SP = r/z with r=1
            cp = c
            # cp = np.cross(c, self.z)
        else:
            raise ValueError('Error, unsupported projection type', self.proj)
        ax = kwargs.get('ax')
        mk = kwargs.get('mk', 'o')
        edge_col = kwargs.get('markeredgecolor', 'k')
        ann = kwargs.get('ann', None)
        lab = kwargs.get('lab', '')
        col = kwargs.get('col', 'k')
        col = col.reshape(1,-1)
        #ax.plot(cp[0], cp[1], linewidth=0, markerfacecolor=col, marker=mk,
        #        markeredgecolor=edge_col, markersize=self.mksize, label=lab)
        mksize = kwargs.get('mksize', self.mksize)
        ax.scatter(cp[0], cp[1], linewidth=0, c=col, marker=mk, edgecolors=edge_col, s=mksize, label=lab)
        # Next 3 lines are necessary in case c_dir[2]=0, as for Euler angles [45, 45, 0]
        if c_dir[2] < 0.000001:
            ax.scatter(-cp[0], -cp[1], linewidth=0, c=col, marker=mk, s=mksize, label=lab)
            #ax.plot(-cp[0], -cp[1], linewidth=0, markerfacecolor=col, marker=mk,
            #        markersize=self.mksize, label=lab)
        if ann:
            ax.annotate(c_dir.view(), (cp[0], cp[1] - 0.1), xycoords='data',
                        fontsize=8, horizontalalignment='center', verticalalignment='center')

    def plot_line_between_crystal_dir(self, c1, c2, ax=None, steps=11, col='k'):
        '''Plot a curve between two crystal directions.

        The curve is actually composed of several straight lines segments to
        draw from direction 1 to direction 2.

        :param c1: vector describing crystal direction 1
        :param c2: vector describing crystal direction 2
        :param ax: a reference to a pyplot ax to draw the line
        :param int steps: number of straight lines composing the curve (11 by default)
        :param col: line color (black by default)
        '''
        path = np.zeros((steps, 2), dtype=float)
        for j, i in enumerate(np.linspace(0., 1., steps)):
            ci = i * c1 + (1 - i) * c2
            ci /= np.linalg.norm(ci)
            if self.proj == 'stereo':
                ci += self.z
                ci /= ci[2]
            path[j, 0] = ci[0]
            path[j, 1] = ci[1]
        ax.plot(path[:, 0], path[:, 1], color=col, markersize=self.mksize, linewidth=2)

    def plot_pf_background(self, ax, labels=True):
        '''Function to plot the background of the pole figure.

        :param ax: a reference to a pyplot ax to draw the backgroud.
        :param bool labels: add lables to axes (True by default).
        '''
        an = np.linspace(0, 2 * np.pi, 100)
        #plt.hold('on')
        ax.plot(np.cos(an), np.sin(an), 'k-')
        ax.plot([-1, 1], [0, 0], 'k-')
        ax.plot([0, 0], [-1, 1], 'k-')
        axe_labels = ['X', 'Y', 'Z']
        if self.axis == 'Z':
            (h, v, u) = (0, 1, 2)
        elif self.axis == 'Y':
            (h, v, u) = (0, 2, 1)
        else:
            (h, v, u) = (1, 2, 0)
        if labels:
            ax.annotate(axe_labels[h], (1.01, 0.0), xycoords='data', fontsize=16,
                        horizontalalignment='left', verticalalignment='center')
            ax.annotate(axe_labels[v], (0.0, 1.01), xycoords='data', fontsize=16,
                        horizontalalignment='center', verticalalignment='bottom')

    def plot_pf_dir(self, c_dir, **kwargs):
        '''Plot a crystal direction in a direct pole figure.'''
        if self.axis == 'Z':
            (h, v, u) = (0, 1, 2)
        elif self.axis == 'Y':
            (h, v, u) = (0, 2, 1)
        else:
            (h, v, u) = (1, 2, 0)
        # the direction to plot is given by c_dir[h,v,u]
        if self.verbose:
            print('corrected for pf axis:', c_dir[[h, v, u]])
        self.plot_crystal_dir(c_dir[[h, v, u]], **kwargs)

    def plot_pf(self, ax=None, mk='o', ann=False):
        '''Create the direct pole figure.

        :param ax: a reference to a pyplot ax to draw the poles.
        :param mk: marker used to plot the poles (disc by default).
        :param bool ann: Annotate the pole with the coordinates of the vector if True (False by default).
        '''
        self.plot_pf_background(ax)
        kwargs = {'ax': ax, 'mk': mk, 'ann': ann}
        if self.resize_markers:
            # compute a list of the grain volume fractions
            fracs = self.microstructure.get_grain_volume_fractions()
            frac_max = max(fracs)
            print('frac max', frac_max)
        for ii, grain in enumerate(self.microstructure.grains):
            g = grain.orientation_matrix()
            gt = g.transpose()
            if self.resize_markers:
                kwargs['mksize'] = 0.15 * np.sqrt(fracs[ii] / frac_max) * 1000
            label = ''
            if self.map_field == 'grain_id':
                label = 'grain ' + str(grain.id)
            kwargs['lab'] = label

            for i, hkl_plane in enumerate(self.poles):
                if i > 0:
                    kwargs['lab'] = ''
                c = hkl_plane.normal()
                c_rot = gt.dot(c)
                if self.verbose:
                    h, k, l = hkl_plane.miller_indices()
                    print('plotting (%d%d%d) with normal %s in sample CS (corrected for pf axis): %s' % (h, k, l, c, c_rot))
                col = self.get_color_from_field(grain)
                kwargs['col'] = col
                self.plot_pf_dir(c_rot, **kwargs)
        ax.axis([-1.1, 1.1, -1.1, 1.1])
        if self.pflegend and self.map_field == 'grain_id':
            ax.legend(bbox_to_anchor=(0.05, 1), loc=1, numpoints=1, prop={'size': 10})
        ax.axis('off')
        ax.set_title('{%s} direct %s projection' % (self.family, self.proj))

    def create_pf_contour(self, ax=None, ang_step=10):
        '''Compute the distribution of orientation and plot it using contouring.

        This plot the distribution of orientation in the microstructure
        associated with this PoleFigure instance, as a continuous
        distribution using angular bining with the specified step.
        the distribution is constructed at runtime by discretizing the
        angular space and counting the number of poles in each bin.
        Then the plot_pf_contour method is called to actually plot the data.

        :param ax: a reference to a pyplot ax to draw the contours.
        :param int ang_step: angular step in degrees to use for constructing the orientation distribution data (10 degrees by default)
        '''
        # discretise the angular space (azimuth and altitude)
        ang_step *= np.pi / 180  # change to radians
        n_phi = int(1 + 2 * np.pi / ang_step)
        n_psi = int(1 + 0.5 * np.pi / ang_step)
        phis = np.linspace(0, 2 * np.pi, n_phi)
        psis = np.linspace(0, np.pi / 2, n_psi)
        xv, yv = np.meshgrid(phis, psis)
        values = np.zeros((n_psi, n_phi), dtype=int)
        for grain in self.microstructure.grains:
            g = grain.orientation_matrix()
            gt = g.transpose()
            for hkl_plane in self.poles:
                c = hkl_plane.normal()
                c_rot = gt.dot(c)
                # handle poles pointing down
                if c_rot[2] < 0:
                    c_rot *= -1  # make unit vector have z>0
                if c_rot[1] >= 0:
                    phi = np.arccos(c_rot[0] / np.sqrt(c_rot[0] ** 2 + c_rot[1] ** 2))
                else:
                    phi = 2 * np.pi - np.arccos(c_rot[0] / np.sqrt(c_rot[0] ** 2 + c_rot[1] ** 2))
                psi = np.arccos(c_rot[2])  # since c_rot is normed
                i_phi = int((phi + 0.5 * ang_step) / ang_step) % n_phi
                j_psi = int((psi + 0.5 * ang_step) / ang_step) % n_psi
                values[j_psi, i_phi] += 1
        if self.proj == 'stereo':  # double check which one is flat/stereo
            x = (2 * yv / np.pi) * np.cos(xv)
            y = (2 * yv / np.pi) * np.sin(xv)
        else:
            x = np.sin(yv) * np.cos(xv)
            y = np.sin(yv) * np.sin(xv)
        # close the pole figure by duplicating azimuth=0
        values[:, -1] = values[:, 0]
        self.plot_pf_contour(ax, x, y, values)

    def plot_pf_contour(self, ax, x, y, values):
        '''Plot the direct pole figure using contours. '''
        self.plot_pf_background(ax)
        ax.contourf(x, y, values)
        # ax.plot(x, y, 'ko')
        ax.axis([-1.1, 1.1, -1.1, 1.1])
        ax.axis('off')
        ax.set_title('{%s} direct %s projection' % (self.family, self.proj))

    def sst_symmetry(self, v):
        """Transform a given vector according to the lattice symmetry associated with the pole figure.

        This function transform a vector so that it lies in the smallest symmetry equivalent zone.

        :param v: the vector to transform.
        :return: the transformed vector.
        """
        # get the symmetry from the lattice associated with the pole figure
        symmetry = self.lattice._symmetry
        if symmetry is Symmetry.cubic:
            return PoleFigure.sst_symmetry_cubic(v)
        elif symmetry is Symmetry.hexagonal:
            syms = symmetry.symmetry_operators()
            for i in range(syms.shape[0]):
                sym = syms[i]
                v_sym = np.dot(sym, v)
                # look at vectors pointing up
                if v_sym[2] < 0:
                    v_sym *= -1
                # now evaluate if projection is in the sst
                if v_sym[1] < 0 or v_sym[0] < 0:
                    continue
                elif v_sym[1] / v_sym[0] > np.tan(np.pi / 6):
                    continue
                else:
                    break
            return v_sym
        else:
            print('unsupported symmetry for the moment: %s' % symmetry)
            return None

    @staticmethod
    def sst_symmetry_cubic(z_rot):
        '''Transform a given vector according to the cubic symmetry.

        This function transform a vector so that it lies in the unit SST triangle.

        :param z_rot: vector to transform.
        :return: the transformed vector.
        '''
        if z_rot[0] < 0: z_rot[0] = -z_rot[0]
        if z_rot[1] < 0: z_rot[1] = -z_rot[1]
        if z_rot[2] < 0: z_rot[2] = -z_rot[2]

        if (z_rot[2] > z_rot[1]):
            z_rot[1], z_rot[2] = z_rot[2], z_rot[1]

        if (z_rot[1] > z_rot[0]):
            z_rot[0], z_rot[1] = z_rot[1], z_rot[0]

        if (z_rot[2] > z_rot[1]):
            z_rot[1], z_rot[2] = z_rot[2], z_rot[1]

        return np.array([z_rot[1], z_rot[2], z_rot[0]])

    def get_color_from_field(self, grain):
        """Get the color of the given grain according to the chosen field.
        
        This function will return the color associated with the given grain. Depending on how the pole figure has been 
        configured (see the `set_map_field` function), it will be obtained from:
         
         * the grain id, according to the `Microstructure.rand_cmap` function
         * ipf the colour will reflect the orientation according to the IPF scheme
         * the field value mapped on a pyplot color map if the lut field of the PoleFigure instance is a string.
         * a color directly read from the lut field; in this case the field value must reflect the category of the given grain. 
        
        :param grain: the `Grain` instance.
        :return: the color as a 3 element numpy array representing the rgb values. 
        """
        if self.map_field:
            if self.map_field == 'grain_id':
                col = Microstructure.rand_cmap().colors[grain.id]
            elif self.map_field == 'ipf':
                if self.axis == 'X':
                    axis = np.array([1., 0., 0.])
                elif self.axis == 'Y':
                    axis = np.array([0., 1., 0.])
                else:
                    axis = np.array([0., 0., 1.])
                col = grain.orientation.get_ipf_colour(axis=axis)
            else:
                # retreive the position of the grain in the list
                rank = self.microstructure.grains.index(grain)
                if type(self.lut) is str:
                    # get the color map from pyplot
                    color_map = cm.get_cmap(self.lut, 256)
                    # use the field value for this grain and the field range bounds
                    color = int(255 * max(min((self.field[rank] - self.field_min_level) / float(
                        self.field_max_level - self.field_min_level), 1.0), 0.0))
                    col = color_map(np.arange(256))[color]
                else:
                    col = self.lut[self.field[rank]]  # directly access the color
            return col
        else:
            return np.array([0., 0., 0.])

    def plot_sst(self, **kwargs):
        """ Create the inverse pole figure in the unit standard triangle.

        :param ax: a reference to a pyplot ax to draw the poles.
        :param mk: marker used to plot the poles (square by default).
        :param bool ann: Annotate the pole with the coordinates of the vector if True (False by default).
        """
        # first draw the boundary of the symmetry domain limited by 3 hkl plane normals, called here A, B and C
        symmetry = self.lattice.get_symmetry()
        ax = kwargs.get('ax')
        if symmetry is Symmetry.cubic:
            sst_poles = [(0, 0, 1), (1, 0, 1), (1, 1, 1)]
            ax.axis([-0.05, 0.45, -0.05, 0.40])
        elif symmetry is Symmetry.hexagonal:
            sst_poles = [(0, 0, 1), (2, -1, 0), (1, 0, 0)]
            ax.axis([-0.05, 1.05, -0.05, 0.6])
        else:
            print('unssuported symmetry: %s' % symmetry)
        A = HklPlane(*sst_poles[0], lattice=self.lattice)
        B = HklPlane(*sst_poles[1], lattice=self.lattice)
        C = HklPlane(*sst_poles[2], lattice=self.lattice)
        self.plot_line_between_crystal_dir(A.normal(), B.normal(), ax=ax, col='k')
        self.plot_line_between_crystal_dir(B.normal(), C.normal(), ax=ax, col='k')
        self.plot_line_between_crystal_dir(C.normal(), A.normal(), ax=ax, col='k')
        # display the 3 crystal axes
        poles = [A, B, C]
        v_align = ['top', 'top', 'bottom']
        for i in range(3):
            hkl = poles[i]
            c_dir = hkl.normal()
            c = c_dir + self.z
            c /= c[2]  # SP'/SP = r/z with r=1
            pole_str = '%d%d%d' % hkl.miller_indices()
            if symmetry is Symmetry.hexagonal:
                pole_str = '%d%d%d%d' % HklPlane.three_to_four_indices(*hkl.miller_indices())
            ax.annotate(pole_str, (c[0], c[1] - (2 * (i < 2) - 1) * 0.01), xycoords='data',
                        fontsize=12, horizontalalignment='center', verticalalignment=v_align[i])

        # now plot the sample axis
        if self.resize_markers:
            # compute a list of the grain volume fractions
            fracs = self.microstructure.get_grain_volume_fractions()
            frac_max = max(fracs)
        for ii, grain in enumerate(self.microstructure.grains):
            g = grain.orientation_matrix()
            if self.resize_markers:
                kwargs['mksize'] = 0.15 * np.sqrt(fracs[ii] / frac_max) * 1000
            # compute axis and apply SST symmetry
            if self.axis == 'Z':
                axis = self.z
            elif self.axis == 'Y':
                axis = self.y
            else:
                axis = self.x
            axis_rot = self.sst_symmetry(g.dot(axis))
            label = ''
            if self.map_field == 'grain_id':
                label = 'grain ' + str(grain.id)
            kwargs['lab'] = label
            kwargs['col'] = self.get_color_from_field(grain)
            self.plot_crystal_dir(axis_rot, **kwargs)
            if self.verbose:
                print('plotting %s in crystal CS: %s' % (self.axis, axis_rot))
        ax.axis('off')
        ax.set_title('%s-axis SST inverse %s projection' % (self.axis, self.proj))

    def plot_ipf_symmetry(self, ax):
        """ Plot the inverse pole figure elements of symmetry.
        
        This assume the symmetry from the `Lattice` associated with the first pole.
        :param ax: a reference to a pyplot ax to draw the poles.
        """
        kwargs = {'ax': ax, 'col': 'k', 'ann': False}
        symmetry = self.lattice.get_symmetry()
        if symmetry is Symmetry.cubic:
            for c in self.c111s:
                for i in range(3):
                    d = c.copy()
                    d[i] = 0
                    e = np.zeros_like(c)
                    e[i] = c[i]
                    self.plot_line_between_crystal_dir(c, d, ax=ax)
                    self.plot_line_between_crystal_dir(c, e, ax=ax)
            markers = ['s', 'o', '^']
            for i, dirs in enumerate([self.c001s, self.c011s, self.c111s]):
                kwargs['mk'] = markers[i]
                [self.plot_crystal_dir(c, **kwargs) for c in dirs]
                # also plot the negative direction of those lying in the plane z==0
                for c in dirs:
                    if np.dot(c, self.z) == 0.0:
                        self.plot_crystal_dir(-c, **kwargs)
        elif symmetry is Symmetry.hexagonal:
            from math import sin, cos, pi
            for angle in np.linspace(0, 2 * pi, 12, endpoint=False):
                ax.plot([0, cos(angle)], [0, sin(angle)], color='k', markersize=self.mksize, linewidth=2)
        else:
            print('skipping unsupported symmetry for now')

    def plot_ipf(self, plot_symmetry=False, **kwargs):
        """ Create the inverse pole figure for direction Z.

        :param bool plot_symmetry: if True plot the lines representing the symmetry of the lattice.
        :param ax: a reference to a pyplot ax to draw the poles.
        :param mk: marker used to plot the poles (square by default).
        :param bool ann: Annotate the pole with the coordinates of the vector if True (False by default).
        """
        ax = kwargs.get('ax')
        self.plot_pf_background(ax, labels=False)
        if plot_symmetry:
            self.plot_ipf_symmetry(ax)
        # now plot the sample axis
        for grain in self.microstructure.grains:
            g = grain.orientation_matrix()
            if self.axis == 'Z':
                axis = self.z
            elif self.axis == 'Y':
                axis = self.y
            else:
                axis = self.x
            axis_rot = g.dot(axis)
            kwargs['col'] = self.get_color_from_field(grain)
            self.plot_crystal_dir(axis_rot, **kwargs)
            if self.verbose:
                print('plotting ', self.axis, ' in crystal CS:', axis_rot)
        ax.axis([-1.1, 1.1, -1.1, 1.1])
        ax.axis('off')
        ax.set_title('%s-axis inverse %s projection' % (self.axis, self.proj))

    @staticmethod
    def plot(orientations, **kwargs):
        '''Plot a pole figure (both direct and inverse) for a list of orientations.

        :param orientations: the list of crystalline :py:class:`~pymicro.crystal.microstructure.Orientation` to plot.
        '''
        micro = Microstructure()
        if isinstance(orientations, list):
            for i in range(len(orientations)):
                micro.grains.append(Grain(i + 1, orientations[i]))
        elif isinstance(orientations, Orientation):
            micro.grains.append(Grain(1, orientations))
        else:
            print('Unrecognized argument: %s' % orientations.__repr__)
        pf = PoleFigure(microstructure=micro, **kwargs)
        pf.plot_pole_figures(display=True)

    @staticmethod
    def plot_euler(phi1, Phi, phi2, **kwargs):
        '''Directly plot a pole figure for a single orientation given its
        three Euler angles.

        ::

          PoleFigure.plot_euler(10, 20, 30)

        :param float phi1: first Euler angle (in degree).
        :param float Phi: second Euler angle (in degree).
        :param float phi2: third Euler angle (in degree).
        '''
        PoleFigure.plot(Orientation.from_euler(np.array([phi1, Phi, phi2])), **kwargs)


class TaylorModel:
    '''A class to carry out texture evolution with the Taylor model.

    Briefly explain the full constrained Taylor model [ref 1938].
    '''

    def __init__(self, microstructure):
        self.micro = microstructure  # Microstructure instance
        self.slip_systems = SlipSystem.get_slip_systems('111')
        self.nact = 5  # number of active slip systems in one grain to accomodate the plastic strain
        self.dt = 1.e-3
        self.max_time = 0.001  # sec
        self.time = 0.0
        self.L = np.array([[-0.5, 0.0, 0.0], [0.0, -0.5, 0.0], [0.0, 0.0, 1.0]])  # velocity gradient

    def compute_step(self, g, check=True):
        Wc = np.zeros((3, 3), dtype=np.float)
        # compute Schmid factors
        SF = []
        for s in self.slip_systems:
            SF.append(g.schmid_factor(s))
        ss_rank = np.zeros(self.nact, dtype=int)
        # rank the slip systems by SF
        for i in range(self.nact):
            ss_rank[i] = np.argmax(SF)
            print('index of ss % d is %d' % (i, ss_rank[i]))
            SF[ss_rank[i]] = 0.0
        # now we need to solve: L = gam1*m1 + gam2*m2+ ...
        iu = np.triu_indices(3)  # indices of the upper part of a 3x3 matrix
        L = self.L[iu][:5]  # form a vector with the velocity gradient components
        M = np.zeros((5, self.nact), dtype=np.float)
        for i in range(len(ss_rank)):
            s = self.slip_systems[ss_rank[i]]
            m = g.orientation.slip_system_orientation_tensor(s)
            # m = g.orientation.slip_system_orientation_strain_tensor(s)
            M[0, i] += m[0, 0]
            M[1, i] += m[0, 1]
            M[2, i] += m[0, 2]
            M[3, i] += m[1, 1]
            M[4, i] += m[1, 2]
            # M[5, i] += m[2, 2]
        dgammas = np.linalg.lstsq(M, L, rcond=1.e-3)[0]
        '''
        U, s, V = np.linalg.svd(M) # solve by SVD
        print('U:\n')
        print(U)
        print('s:\n')
        print(s)
        print('V:\n')
        print(V)
        pinv_svd = np.dot(np.dot(V.T, np.linalg.inv(np.diag(s))), U.T)
        dgammas_svd = np.dot(pinv_svd, L) # solving Ax=b computing x = A^-1*b
        print('dgammas (SVD) =', dgammas_svd)
        '''
        print('dgammas (LST) =', dgammas)
        if check:
            # check consistency
            Lcheck = np.zeros((3, 3), dtype=np.float)
            for i in range(len(ss_rank)):
                s = self.slip_systems[ss_rank[i]]
                ms = g.orientation.slip_system_orientation_tensor(s)
                # ms = g.orientation.slip_system_orientation_strain_tensor(s)
                Lcheck += dgammas[i] * ms
            print('check:', np.sum(Lcheck - self.L), '\n', Lcheck)
            if abs(np.sum(Lcheck - self.L)) > 1e-1:
                raise ValueError(
                    'Problem with solving for plastic slip, trying to increase the number of active slip systems')
        # compute the plastic spin
        for i in range(len(ss_rank)):
            s = self.slip_systems[ss_rank[i]]
            qs = g.orientation.slip_system_orientation_rotation_tensor(s)
            Wc += dgammas[i] * qs
        print('plastic spin:\n', Wc)
        return Wc, dgammas



"""ebsd module to manipulate Electron Back Scattered data sets."""
import h5py
import numpy as np
import os


class OimPhase:
	def __init__(self, id):
		self.id = id
		self.materialName = ''
		self.formula = ''
		self.info = ''
		self.symmetry = 0
		self.latticeConstants =[0] * 6 # a, b, c, alpha, beta, gamma (degrees)
		self.hklFamilies = []
		self.elasticConstants = []
		self.categories = []


class OimHklFamily:
	def __init__(self):
		self.hkl = [0, 0, 0]
		self.useInIndexing = 0
		self.diffractionIntensity = 0.0
		self.showBands = 0


class OimScan:
    """OimScan class to handle files from EDAX software OIM"""

    def __init__(self, shape, resolution=(1.0, 1.0)):
        """Create an empty EBSD scan."""
        self.xstar = 0
        self.ystar = 0
        self.zstar = 0
        self.workingDistance = 0
        self.gridType = 'SqrGrid'
        self.nRows = shape[0]
        self.nColumns = shape[1]
        self.xStep = resolution[0]
        self.yStep = resolution[1]
        self.operator = ''
        self.sampleId = ''
        self.scanId = ''
        self.phase_list = []
        self.init_arrays()

    def init_arrays(self):
        self.euler = np.zeros((self.nRows, self.nColumns, 3))
        self.x = np.zeros((self.nRows, self.nColumns))
        self.y = np.zeros((self.nRows, self.nColumns))
        self.iq = np.zeros((self.nRows, self.nColumns))
        self.ci = np.zeros((self.nRows, self.nColumns))
        self.phase = np.zeros((self.nRows, self.nColumns), dtype='int')

    @staticmethod
    def from_file(file_path):
        """Create a new EBSD scan by reading a data file. At present, only hdf5 format is supported."""
        base_name, ext = os.path.splitext(os.path.basename(file_path))
        print(base_name, ext)
        if ext not in ['.h5', '.hdf5']:
            raise ValueError('only HDF5 format is supported, please convert your scan')
        scan = OimScan((0, 0))
        with h5py.File(file_path, 'r') as f:
            # find out the scan key (the third one)
            key_list = [key for key in f.keys()]
            scan_key = key_list[2]
            print('reading EBSD scan %s from file %s' % (scan_key, file_path))
            header = f[scan_key]['EBSD']['Header']
            scan.read_header(header)
            # now initialize the fields
            scan.init_arrays()
            data = f[scan_key]['EBSD']['Data']
            scan.read_data(data)
        return scan

    def read_header(self, header):
        # read the header, it contains the following keys: 'Camera Azimuthal Angle', 'Camera Elevation Angle',
        # 'Coordinate System', 'Grid Type', 'Notes', 'Operator', 'Pattern Center Calibration', 'Phase', 'Sample ID',
        # 'Sample Tilt', 'Scan ID', 'Step X', 'Step Y', 'Working Distance', 'nColumns', 'nRows'
        self.xstar = header['Pattern Center Calibration']['x-star'][0]
        self.ystar = header['Pattern Center Calibration']['y-star'][0]
        self.zstar = header['Pattern Center Calibration']['z-star'][0]
        self.workingDistance = header['Camera Elevation Angle'][0]
        self.gridType = header['Grid Type'][0].decode('utf-8')
        if self.gridType != 'SqrGrid':
            raise ValueError('only square grid is supported, please convert your scan')
        self.nColumns = header['nColumns'][0]
        self.nRows = header['nRows'][0]
        self.xStep = header['Step X'][0]
        self.yStep = header['Step Y'][0]
        self.operator = header['Operator'][0].decode('utf-8')
        self.sampleId = header['Sample ID'][0].decode('utf-8')
        self.scanId = header['Scan ID'][0].decode('utf-8')
        # get the different phases
        for key in header['Phase'].keys():
            phase = header['Phase'][key]
            # each phase has the following keys: 'Formula', 'Info', 'Lattice Constant a', 'Lattice Constant alpha',
            # 'Lattice Constant b', 'Lattice Constant beta', 'Lattice Constant c', 'Lattice Constant gamma',
            # 'Laue Group', 'MaterialName', 'NumberFamilies', 'Point Group', 'Symmetry', 'hkl Families'
            phase = OimPhase(int(key))
            phase.materialName = header['Phase'][key]['MaterialName'][0].decode('utf-8')
            phase.formula = header['Phase'][key]['Formula'][0].decode('utf-8')
            phase.info = header['Phase'][key]['Info'][0].decode('utf-8')
            phase.symmetry = header['Phase'][key]['Symmetry'][0]
            phase.latticeConstants[0] = header['Phase'][key]['Lattice Constant a'][0]
            phase.latticeConstants[1] = header['Phase'][key]['Lattice Constant b'][0]
            phase.latticeConstants[2] = header['Phase'][key]['Lattice Constant c'][0]
            phase.latticeConstants[3] = header['Phase'][key]['Lattice Constant alpha'][0]
            phase.latticeConstants[4] = header['Phase'][key]['Lattice Constant beta'][0]
            phase.latticeConstants[5] = header['Phase'][key]['Lattice Constant gamma'][0]
            for row in header['Phase'][key]['hkl Families']:
                family = OimHklFamily()
                family.hkl = [row[0], row[1], row[2]]
                family.useInIndexing = row[4]
                family.diffractionIntensity = row[3]
                family.showBands = row[5]
                phase.hklFamilies.append(family)
            phase.elasticConstants = [[0, 0, 0, 0, 0, 0] * 6 for i in range(6)]
            phase.categories = [0, 0, 0, 0, 0]
            self.phase_list.append(phase)

    def read_data(self, data):
        self.euler[:, :, 0] = np.reshape(data['Phi1'], (self.nRows, self.nColumns))
        self.euler[:, :, 1] = np.reshape(data['Phi'], (self.nRows, self.nColumns))
        self.euler[:, :, 2] = np.reshape(data['Phi2'], (self.nRows, self.nColumns))
        self.x = np.reshape(data['X Position'], (self.nRows, self.nColumns))
        self.y = np.reshape(data['Y Position'], (self.nRows, self.nColumns))
        self.iq = np.reshape(data['IQ'], (self.nRows, self.nColumns))
        self.ci = np.reshape(data['CI'], (self.nRows, self.nColumns))
        self.phase = np.reshape(data['Phase'], (self.nRows, self.nColumns))

    def to_h5(self, file_name):
        """Write the EBSD scan as a hdf5 file compatible OIM software (in progress)."""
        f = h5py.File('%s.h5' % file_name, 'w')
        f.attrs[' Manufacturer'] = np.string_('EDAX')
        f.attrs[' Version'] = np.string_('OIM Analysis 7.3.0 x64  [09-01-15]')
        # create the group containing the data
        data_container = f.create_group('DataContainer')
        ebsd = data_container.create_group('EBSD')
        ebsd_header = ebsd.create_group('Header')
        ebsd_header.create_dataset('Camera Azimuthal Angle', data=np.array([0.0], dtype=np.float32))
        ebsd_header.create_dataset('Camera Elevation Angle', data=np.array([self.workingDistance], dtype=np.float32))
        pattern_center = ebsd_header.create_group('Pattern Center Calibration')
        pattern_center.create_dataset('x-star', data=np.array(self.xstar, dtype=np.float32))
        pattern_center.create_dataset('y-star', data=np.array(self.ystar, dtype=np.float32))
        pattern_center.create_dataset('z-star', data=np.array(self.zstar, dtype=np.float32))
        ebsd_data = ebsd.create_group('Data')
        ci = ebsd_data.create_dataset('CI', data=self.ci)
        iq = ebsd_data.create_dataset('IQ', data=self.iq)
        phase = ebsd_data.create_dataset('Phase', data=self.phase)
        phi1 = ebsd_data.create_dataset('Phi1', data=self.euler[:, :, 0])
        phi = ebsd_data.create_dataset('Phi', data=self.euler[:, :, 1])
        phi2 = ebsd_data.create_dataset('Phi2', data=self.euler[:, :, 2])
        x = ebsd_data.create_dataset('X Position', data=self.x)
        y = ebsd_data.create_dataset('Y Position', data=self.y)
        f.close()
